from collections.abc import Callable, Iterable, Set
from datetime import datetime, timezone
from pathlib import Path
from typing import cast

from cognite.client import CogniteClient
from cognite.client.data_classes import Asset, AssetFilter, AssetList
from rdflib import RDF, Literal, Namespace

from cognite.neat._graph.models import Triple

from ._base import DEFAULT_SKIP_METADATA_VALUES, ClassicCDFBaseExtractor, InstanceIdPrefix
from ._labels import LabelsExtractor


class AssetsExtractor(ClassicCDFBaseExtractor[Asset]):
    """Extract data from Cognite Data Fusions Assets into Neat.

    Args:
        items (Iterable[Asset]): An iterable of assets.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
        to_type (Callable[[Asset], str | None], optional): A function to convert an asset to a type. Defaults to None.
            If None or if the function returns None, the asset will be set to the default type "Asset".
        total (int, optional): The total number of assets to load. If passed, you will get a progress bar if rich
            is installed. Defaults to None.
        limit (int, optional): The maximal number of assets to load. Defaults to None. This is typically used for
            testing setup of the extractor. For example, if you are extracting 100 000 assets, you might want to
            limit the extraction to 1000 assets to test the setup.
        unpack_metadata (bool, optional): Whether to unpack metadata. Defaults to False, which yields the metadata as
            a JSON string.
        skip_metadata_values (set[str] | frozenset[str] | None, optional): A set of values to skip when unpacking
            metadata. Defaults to frozenset({"nan", "null", "none", ""}).
    """

    _default_rdf_type = "Asset"

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
        to_type: Callable[[Asset], str | None] | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
    ):
        total = client.assets.aggregate_count(filter=AssetFilter(data_set_ids=[{"externalId": data_set_external_id}]))

        return cls(
            client.assets(data_set_external_ids=data_set_external_id),
            namespace,
            to_type,
            total,
            limit,
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
        )

    @classmethod
    def from_hierarchy(
        cls,
        client: CogniteClient,
        root_asset_external_id: str,
        namespace: Namespace | None = None,
        to_type: Callable[[Asset], str | None] | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
    ):
        total = client.assets.aggregate_count(
            filter=AssetFilter(asset_subtree_ids=[{"externalId": root_asset_external_id}])
        )

        return cls(
            cast(
                Iterable[Asset],
                client.assets(asset_subtree_external_ids=root_asset_external_id),
            ),
            namespace,
            to_type,
            total,
            limit,
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
        )

    @classmethod
    def from_file(
        cls,
        file_path: str,
        namespace: Namespace | None = None,
        to_type: Callable[[Asset], str] | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
    ):
        assets = AssetList.load(Path(file_path).read_text())
        return cls(
            assets,
            namespace,
            to_type,
            total=len(assets),
            limit=limit,
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
        )

    def _item2triples(self, asset: Asset) -> list[Triple]:
        """Converts an asset to triples."""
        id_ = self.namespace[f"{InstanceIdPrefix.asset}{asset.id}"]

        type_ = self._get_rdf_type(asset)

        triples: list[Triple] = [(id_, RDF.type, self.namespace[type_])]

        # Create attributes
        if asset.name:
            triples.append((id_, self.namespace.name, Literal(asset.name)))

        if asset.description:
            triples.append((id_, self.namespace.description, Literal(asset.description)))

        if asset.external_id:
            triples.append((id_, self.namespace.external_id, Literal(asset.external_id)))

        if asset.source:
            triples.append((id_, self.namespace.source, Literal(asset.source)))

        # properties' ref creation and update
        triples.append(
            (
                id_,
                self.namespace.created_time,
                Literal(datetime.fromtimestamp(asset.created_time / 1000, timezone.utc)),
            )
        )
        triples.append(
            (
                id_,
                self.namespace.last_updated_time,
                Literal(datetime.fromtimestamp(asset.last_updated_time / 1000, timezone.utc)),
            )
        )

        if asset.labels:
            for label in asset.labels:
                # external_id can create ill-formed URIs, so we create websafe URIs
                # since labels do not have internal ids, we use the external_id as the id
                triples.append(
                    (
                        id_,
                        self.namespace.label,
                        self.namespace[f"{InstanceIdPrefix.label}{LabelsExtractor._label_id(label)}"],
                    )
                )

        if asset.metadata:
            triples.extend(self._metadata_to_triples(id_, asset.metadata))

        # Create connections:
        if asset.parent_id:
            triples.append((id_, self.namespace.parent, self.namespace[f"{InstanceIdPrefix.asset}{asset.parent_id}"]))

        if asset.root_id:
            triples.append((id_, self.namespace.root, self.namespace[f"{InstanceIdPrefix.asset}{asset.root_id}"]))

        if asset.data_set_id:
            triples.append(
                (
                    id_,
                    self.namespace.dataset,
                    self.namespace[f"{InstanceIdPrefix.data_set}{asset.data_set_id}"],
                )
            )

        return triples



from collections.abc import Iterable

from cognite.neat._issues.errors import NeatValueError
from cognite.neat._rules._shared import InputRules, MaybeRules, VerifiedRules
from cognite.neat._rules.importers import BaseImporter
from cognite.neat._rules.models import VERIFIED_RULES_BY_ROLE, RoleTypes

from ._base import RulesPipeline, RulesTransformer
from ._converters import ConvertToRules
from ._verification import VerifyAnyRules


class ImporterPipeline(RulesPipeline[InputRules, VerifiedRules]):
    """This is a standard pipeline that verifies, convert and return the rules from the importer."""

    def __init__(
        self,
        importer: BaseImporter[InputRules],
        items: Iterable[RulesTransformer[InputRules, VerifiedRules]] | None = None,
    ) -> None:
        super().__init__(items or [])
        self._importer = importer

    @classmethod
    def _create_pipeline(cls, importer: BaseImporter[InputRules], role: RoleTypes | None = None) -> "ImporterPipeline":
        items: list[RulesTransformer] = [VerifyAnyRules(errors="continue")]
        if role is not None:
            out_cls = VERIFIED_RULES_BY_ROLE[role]
            items.append(ConvertToRules(out_cls))
        return cls(importer, items)

    @classmethod
    def try_verify(cls, importer: BaseImporter, role: RoleTypes | None = None) -> MaybeRules[VerifiedRules]:
        """This is a standard pipeline that verifies, convert and return the rules from the importer.

        Args:
            importer: The importer to use.
            role: The type of rules to convert to. If None, the rules will not be converted.

        Returns:
            The verified rules.
        """
        return cls._create_pipeline(importer, role).try_execute()

    @classmethod
    def verify(cls, importer: BaseImporter, role: RoleTypes | None = None) -> VerifiedRules:
        """Verify the rules."""
        return cls._create_pipeline(importer, role).execute()

    def try_execute(self) -> MaybeRules[VerifiedRules]:
        """Try to execute the pipeline from importer to rules."""
        rules = self._importer.to_rules()
        return self.try_transform(rules)

    def execute(self) -> VerifiedRules:
        """Execute the pipeline from importer to rules."""
        rules = self._importer.to_rules()
        out = self.transform(rules)
        if isinstance(out, MaybeRules) and out.rules is None:
            raise out.issues.as_errors("Failed to convert rules")

        rules = out.get_rules()
        if rules is None:
            raise NeatValueError("Rules is missing cannot convert")
        return rules



import re
import warnings
from abc import ABC, abstractmethod
from collections import Counter, defaultdict
from collections.abc import Collection, Mapping
from datetime import date, datetime
from typing import Literal, TypeVar, cast

from cognite.client.data_classes import data_modeling as dms
from cognite.client.data_classes.data_modeling import DataModelId, DataModelIdentifier, ViewId
from rdflib import Namespace

from cognite.neat._constants import COGNITE_MODELS, DMS_CONTAINER_PROPERTY_SIZE_LIMIT
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._issues.warnings._models import (
    EnterpriseModelNotBuildOnTopOfCDMWarning,
    SolutionModelBuildOnTopOfCDMWarning,
)
from cognite.neat._issues.warnings.user_modeling import ParentInDifferentSpaceWarning
from cognite.neat._rules._constants import EntityTypes
from cognite.neat._rules._shared import InputRules, JustRules, OutRules, VerifiedRules
from cognite.neat._rules.analysis import DMSAnalysis
from cognite.neat._rules.models import (
    AssetRules,
    DMSInputRules,
    DMSRules,
    DomainRules,
    ExtensionCategory,
    InformationRules,
    SchemaCompleteness,
    SheetList,
    data_types,
)
from cognite.neat._rules.models.data_types import DataType, String
from cognite.neat._rules.models.dms import DMSMetadata, DMSProperty, DMSView
from cognite.neat._rules.models.dms._rules import DMSContainer
from cognite.neat._rules.models.entities import (
    AssetEntity,
    AssetFields,
    ClassEntity,
    ContainerEntity,
    DMSUnknownEntity,
    EdgeEntity,
    MultiValueTypeInfo,
    ReferenceEntity,
    RelationshipEntity,
    ReverseConnectionEntity,
    UnknownEntity,
    ViewEntity,
)
from cognite.neat._rules.models.information import InformationClass, InformationMetadata, InformationProperty
from cognite.neat._rules.models.information._rules_input import (
    InformationInputClass,
    InformationInputProperty,
    InformationInputRules,
)
from cognite.neat._utils.collection_ import remove_list_elements
from cognite.neat._utils.text import to_camel

from ._base import RulesTransformer

T_VerifiedInRules = TypeVar("T_VerifiedInRules", bound=VerifiedRules)
T_VerifiedOutRules = TypeVar("T_VerifiedOutRules", bound=VerifiedRules)
T_InputInRules = TypeVar("T_InputInRules", bound=InputRules)
T_InputOutRules = TypeVar("T_InputOutRules", bound=InputRules)


class ConversionTransformer(RulesTransformer[T_VerifiedInRules, T_VerifiedOutRules], ABC):
    """Base class for all conversion transformers."""

    def transform(self, rules: T_VerifiedInRules | OutRules[T_VerifiedInRules]) -> JustRules[T_VerifiedOutRules]:
        out = self._transform(self._to_rules(rules))
        return JustRules(out)

    @abstractmethod
    def _transform(self, rules: T_VerifiedInRules) -> T_VerifiedOutRules:
        raise NotImplementedError()


class ToCompliantEntities(RulesTransformer[InformationInputRules, InformationInputRules]):  # type: ignore[misc]
    """Converts input rules to rules with compliant entity IDs that match regex patters used
    by DMS schema components."""

    def transform(
        self, rules: InformationInputRules | OutRules[InformationInputRules]
    ) -> OutRules[InformationInputRules]:
        return JustRules(self._transform(self._to_rules(rules)))

    def _transform(self, rules: InformationInputRules) -> InformationInputRules:
        rules.metadata.prefix = self._fix_entity(rules.metadata.prefix)
        rules.classes = self._fix_classes(rules.classes)
        rules.properties = self._fix_properties(rules.properties)
        return rules

    @classmethod
    def _fix_entity(cls, entity: str) -> str:
        entity = re.sub(r"[^_a-zA-Z0-9]+", "_", entity)

        # entity id must start with a letter
        if not entity[0].isalpha():
            entity = "prefix_" + entity
        # and end with a letter or number
        if not entity[-1].isalnum():
            entity = entity + "_suffix"

        # removing any double underscores that could occur
        return re.sub(r"[^a-zA-Z0-9]+", "_", entity)

    @classmethod
    def _fix_class(cls, class_: str | ClassEntity) -> str | ClassEntity:
        if isinstance(class_, str):
            if len(class_.split(":")) == 2:
                prefix, suffix = class_.split(":")
                class_ = f"{cls._fix_entity(prefix)}:{cls._fix_entity(suffix)}"

            else:
                class_ = cls._fix_entity(class_)

        elif isinstance(class_, ClassEntity) and type(class_.prefix) is str:
            class_ = ClassEntity(
                prefix=cls._fix_entity(class_.prefix),
                suffix=cls._fix_entity(class_.suffix),
            )

        return class_

    @classmethod
    def _fix_value_type(
        cls, value_type: str | DataType | ClassEntity | MultiValueTypeInfo
    ) -> str | DataType | ClassEntity | MultiValueTypeInfo:
        fixed_value_type: str | DataType | ClassEntity | MultiValueTypeInfo

        if isinstance(value_type, str):
            # this is a multi value type but as string
            if " | " in value_type:
                value_types = value_type.split(" | ")
                fixed_value_type = " | ".join([cast(str, cls._fix_value_type(v)) for v in value_types])
            # this is value type specified with prefix:suffix string
            elif ":" in value_type:
                fixed_value_type = cls._fix_class(value_type)

            # this is value type specified as suffix only
            else:
                fixed_value_type = cls._fix_entity(value_type)

        # value type specified as instances of DataType, ClassEntity or MultiValueTypeInfo
        elif isinstance(value_type, MultiValueTypeInfo):
            fixed_value_type = MultiValueTypeInfo(
                types=[cast(DataType | ClassEntity, cls._fix_value_type(type_)) for type_ in value_type.types],
            )
        elif isinstance(value_type, ClassEntity):
            fixed_value_type = cls._fix_class(value_type)

        # this is a DataType instance but also we should default to original value
        else:
            fixed_value_type = value_type

        return fixed_value_type

    @classmethod
    def _fix_classes(cls, definitions: list[InformationInputClass]) -> list[InformationInputClass]:
        fixed_definitions = []
        for definition in definitions:
            definition.class_ = cls._fix_class(definition.class_)
            fixed_definitions.append(definition)
        return fixed_definitions

    @classmethod
    def _fix_properties(cls, definitions: list[InformationInputProperty]) -> list[InformationInputProperty]:
        fixed_definitions = []
        for definition in definitions:
            definition.class_ = cls._fix_class(definition.class_)
            definition.property_ = cls._fix_entity(definition.property_)
            definition.value_type = cls._fix_value_type(definition.value_type)
            fixed_definitions.append(definition)
        return fixed_definitions


class InformationToDMS(ConversionTransformer[InformationRules, DMSRules]):
    """Converts InformationRules to DMSRules."""

    def __init__(self, ignore_undefined_value_types: bool = False):
        self.ignore_undefined_value_types = ignore_undefined_value_types

    def _transform(self, rules: InformationRules) -> DMSRules:
        return _InformationRulesConverter(rules).as_dms_rules(self.ignore_undefined_value_types)


class InformationToAsset(ConversionTransformer[InformationRules, AssetRules]):
    """Converts InformationRules to AssetRules."""

    def _transform(self, rules: InformationRules) -> AssetRules:
        return _InformationRulesConverter(rules).as_asset_architect_rules()


class AssetToInformation(ConversionTransformer[AssetRules, InformationRules]):
    """Converts AssetRules to InformationRules."""

    def _transform(self, rules: AssetRules) -> InformationRules:
        return InformationRules.model_validate(rules.model_dump())


class DMSToInformation(ConversionTransformer[DMSRules, InformationRules]):
    """Converts DMSRules to InformationRules."""

    def _transform(self, rules: DMSRules) -> InformationRules:
        return _DMSRulesConverter(rules).as_information_rules()


class ConvertToRules(ConversionTransformer[VerifiedRules, VerifiedRules]):
    """Converts any rules to any rules."""

    def __init__(self, out_cls: type[VerifiedRules]):
        self._out_cls = out_cls

    def _transform(self, rules: VerifiedRules) -> VerifiedRules:
        if isinstance(rules, self._out_cls):
            return rules
        if isinstance(rules, InformationRules) and self._out_cls is DMSRules:
            return InformationToDMS().transform(rules).rules
        if isinstance(rules, InformationRules) and self._out_cls is AssetRules:
            return InformationToAsset().transform(rules).rules
        if isinstance(rules, AssetRules) and self._out_cls is InformationRules:
            return AssetToInformation().transform(rules).rules
        if isinstance(rules, AssetRules) and self._out_cls is DMSRules:
            return InformationToDMS().transform(AssetToInformation().transform(rules)).rules
        if isinstance(rules, DMSRules) and self._out_cls is InformationRules:
            return DMSToInformation().transform(rules).rules
        if isinstance(rules, DMSRules) and self._out_cls is AssetRules:
            return InformationToAsset().transform(DMSToInformation().transform(rules)).rules
        raise ValueError(f"Unsupported conversion from {type(rules)} to {self._out_cls}")


_T_Entity = TypeVar("_T_Entity", bound=ClassEntity | ViewEntity)


class SetIDDMSModel(RulesTransformer[DMSRules, DMSRules]):
    def __init__(self, new_id: DataModelId | tuple[str, str, str]):
        self.new_id = DataModelId.load(new_id)

    def transform(self, rules: DMSRules | OutRules[DMSRules]) -> JustRules[DMSRules]:
        if self.new_id.version is None:
            raise NeatValueError("Version is required when setting a new Data Model ID")
        dump = self._to_rules(rules).dump()
        dump["metadata"]["space"] = self.new_id.space
        dump["metadata"]["external_id"] = self.new_id.external_id
        dump["metadata"]["version"] = self.new_id.version
        # Serialize and deserialize to set the new space and external_id
        # as the default values for the new model.
        return JustRules(DMSRules.model_validate(DMSInputRules.load(dump).dump()))


class ToExtension(RulesTransformer[DMSRules, DMSRules]):
    def __init__(
        self,
        new_model_id: DataModelIdentifier,
        org_name: str = "My",
        type_: Literal["enterprise", "solution"] = "enterprise",
        mode: Literal["read", "write"] = "read",
        dummy_property: str = "GUID",
    ):
        self.new_model_id = DataModelId.load(new_model_id)
        if not self.new_model_id.version:
            raise NeatValueError("Version is required for the new model.")

        self.org_name = org_name
        self.mode = mode
        self.type_ = type_
        self.dummy_property = dummy_property

    def transform(self, rules: DMSRules | OutRules[DMSRules]) -> JustRules[DMSRules]:
        # Copy to ensure immutability
        reference_model = self._to_rules(rules)
        reference_model_id = reference_model.metadata.as_data_model_id()

        # if model is solution then we need to get correct space for views and containers
        if self.type_ == "solution":
            if self.mode not in ["read", "write"]:
                raise NeatValueError(f"Unsupported mode: {self.mode}")

            if reference_model_id in COGNITE_MODELS:
                warnings.warn(
                    SolutionModelBuildOnTopOfCDMWarning(reference_model_id=reference_model_id),
                    stacklevel=2,
                )

            return self._to_solution(reference_model)

        elif self.type_ == "enterprise":
            if reference_model_id not in COGNITE_MODELS:
                warnings.warn(
                    EnterpriseModelNotBuildOnTopOfCDMWarning(reference_model_id=reference_model_id).as_message(),
                    stacklevel=2,
                )

            return self._to_enterprise(reference_model)

        else:
            raise NeatValueError(f"Unsupported data model type: {self.type_}")

    def _has_views_in_multiple_space(self, rules: DMSRules) -> bool:
        return any(view.view.space != rules.metadata.space for view in rules.views)

    def _to_solution(self, reference_rules: DMSRules) -> JustRules[DMSRules]:
        """For creation of solution data model / rules specifically for mapping over existing containers."""

        dump = reference_rules.dump()

        # Prepare new model metadata prior validation
        dump["metadata"]["schema_"] = SchemaCompleteness.partial.value
        dump["metadata"]["data_model_type"] = self.type_
        dump["metadata"]["name"] = f"{self.org_name} {self.type_} data model"
        dump["metadata"]["space"] = self.new_model_id.space
        dump["metadata"]["external_id"] = self.new_model_id.external_id
        dump["metadata"]["version"] = self.new_model_id.version

        # dropping reference and last from the dump as they can cause validation
        # issues especially if reference is enterprise model build on top of CDM
        dump.pop("reference", None)
        dump.pop("last", None)

        # Set implement to NONE for all views
        for view in dump["views"]:
            view["implements"] = None

        if self._has_views_in_multiple_space(reference_rules):
            views_to_remove = []
            for view in dump["views"]:
                if ":" in view["view"]:
                    views_to_remove.append(view)

            dump["views"] = remove_list_elements(dump["views"], views_to_remove)

        solution_model = DMSRules.model_validate(DMSInputRules.load(dump).dump())

        # Dropping containers coming from reference model
        solution_model.containers = SheetList[DMSContainer]()

        # We want to map properties to existing containers allowing extension
        for prop in solution_model.properties:
            if prop.container and prop.container.space == self.new_model_id.space:
                prop.container = ContainerEntity(
                    space=reference_rules.metadata.space,
                    externalId=prop.container.suffix,
                )

        # If reference model on which we are mapping one of Cognite Data Models
        # since we want to affix these with the organization name
        if reference_rules.metadata.as_data_model_id() in COGNITE_MODELS:
            # Remove Cognite affix in view external_id / suffix.
            for prop in solution_model.properties:
                prop.view = self._remove_cognite_affix(prop.view)
                prop.class_ = self._remove_cognite_affix(prop.class_)
                if isinstance(prop.value_type, ViewEntity):
                    prop.value_type = self._remove_cognite_affix(prop.value_type)
            for view in solution_model.views:
                view.view = self._remove_cognite_affix(view.view)
                view.class_ = self._remove_cognite_affix(view.class_)

        if self.mode == "write":
            _, new_containers, new_properties = self._get_new_components(solution_model)

            # Here we add ONLY dummy properties of the solution model and
            # corresponding solution model space containers to hold them
            solution_model.containers.extend(new_containers)
            solution_model.properties.extend(new_properties)

        return JustRules(solution_model)

    def _to_enterprise(self, reference_model: DMSRules) -> JustRules[DMSRules]:
        dump = reference_model.dump()

        # This is must prior model validation to avoid validation issues
        dump["metadata"]["schema_"] = SchemaCompleteness.partial.value

        # This will create reference model components in the enterprise model space
        enterprise_model = DMSRules.model_validate(DMSInputRules.load(dump).dump())

        # Post validation metadata update:
        enterprise_model.metadata.name = self.type_
        enterprise_model.metadata.name = f"{self.org_name} {self.type_} data model"
        enterprise_model.metadata.space = self.new_model_id.space
        enterprise_model.metadata.external_id = self.new_model_id.external_id
        enterprise_model.metadata.version = cast(str, self.new_model_id.version)

        if reference_model.metadata.as_data_model_id() in COGNITE_MODELS:
            enterprise_model.reference = reference_model

        # Here we are creating enterprise specific components
        enterprise_views, enterprise_containers, enterprise_properties = self._get_new_components(enterprise_model)

        # And we are adding them to the enterprise model
        # extending reference views with new ones
        enterprise_model.views.extend(enterprise_views)

        # while overwriting containers and properties with new ones
        enterprise_model.containers = enterprise_containers
        enterprise_model.properties = enterprise_properties

        return JustRules(enterprise_model)

    def _remove_cognite_affix(self, entity: _T_Entity) -> _T_Entity:
        """This method removes `Cognite` affix from the entity."""
        new_suffix = entity.suffix.replace("Cognite", self.org_name or "")
        if isinstance(entity, ViewEntity):
            return ViewEntity(space=entity.space, externalId=new_suffix, version=entity.version)  # type: ignore[return-value]
        elif isinstance(entity, ClassEntity):
            return ClassEntity(prefix=entity.prefix, suffix=new_suffix, version=entity.version)  # type: ignore[return-value]
        raise ValueError(f"Unsupported entity type: {type(entity)}")

    def _get_new_components(
        self, rules: DMSRules
    ) -> tuple[SheetList[DMSView], SheetList[DMSContainer], SheetList[DMSProperty]]:
        new_views = SheetList[DMSView]()
        new_containers = SheetList[DMSContainer]()
        new_properties = SheetList[DMSProperty]()

        for definition in rules.views:
            view_entity = self._remove_cognite_affix(definition.view)

            view_entity.version = cast(str, self.new_model_id.version)
            view_entity.prefix = self.new_model_id.space
            container_entity = ContainerEntity(space=view_entity.prefix, externalId=view_entity.external_id)
            class_entity = ClassEntity(prefix=view_entity.prefix, suffix=view_entity.suffix)

            view = DMSView(
                view=view_entity,
                implements=[definition.view],
                in_model=True,
                class_=class_entity,
                name=definition.name,
            )

            container = DMSContainer(
                container=container_entity,
                class_=class_entity,
            )

            property_ = DMSProperty(
                view=view_entity,
                view_property=f"{to_camel(view_entity.suffix)}{self.dummy_property}",
                value_type=String(),
                nullable=True,
                immutable=False,
                is_list=False,
                container=container_entity,
                container_property=f"{to_camel(view_entity.suffix)}{self.dummy_property}",
                class_=class_entity,
                property_=f"{to_camel(view_entity.suffix)}{self.dummy_property}",
            )

            new_properties.append(property_)
            new_views.append(view)
            new_containers.append(container)

        return new_views, new_containers, new_properties


class ReduceCogniteModel(RulesTransformer[DMSRules, DMSRules]):
    _ASSET_VIEW = ViewId("cdf_cdm", "CogniteAsset", "v1")
    _VIEW_BY_COLLECTION: Mapping[Literal["3D", "Annotation", "BaseViews"], frozenset[ViewId]] = {
        "3D": frozenset(
            {
                ViewId("cdf_cdm", "Cognite3DModel", "v1"),
                ViewId("cdf_cdm", "Cognite3DObject", "v1"),
                ViewId("cdf_cdm", "Cognite3DRevision", "v1"),
                ViewId("cdf_cdm", "Cognite3DTransformation", "v1"),
                ViewId("cdf_cdm", "Cognite360Image", "v1"),
                ViewId("cdf_cdm", "Cognite360ImageAnnotation", "v1"),
                ViewId("cdf_cdm", "Cognite360ImageCollection", "v1"),
                ViewId("cdf_cdm", "Cognite360ImageModel", "v1"),
                ViewId("cdf_cdm", "Cognite360ImageStation", "v1"),
                ViewId("cdf_cdm", "CogniteCADModel", "v1"),
                ViewId("cdf_cdm", "CogniteCADNode", "v1"),
                ViewId("cdf_cdm", "CogniteCADRevision", "v1"),
                ViewId("cdf_cdm", "CogniteCubeMap", "v1"),
                ViewId("cdf_cdm", "CognitePointCloudModel", "v1"),
                ViewId("cdf_cdm", "CognitePointCloudRevision", "v1"),
                ViewId("cdf_cdm", "CognitePointCloudVolume", "v1"),
            }
        ),
        "Annotation": frozenset(
            {
                ViewId("cdf_cdm", "CogniteAnnotation", "v1"),
                ViewId("cdf_cdm", "CogniteDiagramAnnotation", "v1"),
            }
        ),
        "BaseViews": frozenset(
            {
                ViewId("cdf_cdm", "CogniteDescribable", "v1"),
                ViewId("cdf_cdm", "CogniteSchedulable", "v1"),
                ViewId("cdf_cdm", "CogniteSourceable", "v1"),
                ViewId("cdf_cdm", "CogniteVisualizable", "v1"),
            }
        ),
    }

    def __init__(self, drop: Collection[Literal["3D", "Annotation", "BaseViews"] | str]):
        self.drop_collection = cast(
            list[Literal["3D", "Annotation", "BaseViews"]],
            [collection for collection in drop if collection in self._VIEW_BY_COLLECTION],
        )
        self.drop_external_ids = {external_id for external_id in drop if external_id not in self._VIEW_BY_COLLECTION}

    def transform(self, rules: DMSRules | OutRules[DMSRules]) -> JustRules[DMSRules]:
        verified = self._to_rules(rules)
        if verified.metadata.as_data_model_id() not in COGNITE_MODELS:
            raise NeatValueError(f"Can only reduce Cognite Data Models, not {verified.metadata.as_data_model_id()}")

        exclude_views = {view for collection in self.drop_collection for view in self._VIEW_BY_COLLECTION[collection]}
        exclude_views |= {view.view.as_id() for view in verified.views if view.view.suffix in self.drop_external_ids}
        new_model = verified.model_copy(deep=True)

        properties_by_view = DMSAnalysis(new_model).classes_with_properties(consider_inheritance=True)

        new_model.views = SheetList[DMSView](
            [view for view in new_model.views if view.view.as_id() not in exclude_views]
        )
        new_properties = SheetList[DMSProperty]()
        for view in new_model.views:
            for prop in properties_by_view[view.view]:
                if self._is_asset_3D_property(prop):
                    # We filter out the 3D property of asset
                    continue
                new_properties.append(prop)
        new_model.properties = new_properties

        return JustRules(new_model)

    def _is_asset_3D_property(self, prop: DMSProperty) -> bool:
        if "3D" not in self.drop_collection:
            return False
        return prop.view.as_id() == self._ASSET_VIEW and prop.property_ == "object3D"


class _InformationRulesConverter:
    def __init__(self, information: InformationRules):
        self.rules = information
        self.is_addition = (
            self.rules.metadata.schema_ is SchemaCompleteness.extended
            and self.rules.metadata.extension is ExtensionCategory.addition
        )
        self.is_reshape = (
            self.rules.metadata.schema_ is SchemaCompleteness.extended
            and self.rules.metadata.extension is ExtensionCategory.reshape
        )
        if self.rules.last:
            self.last_classes = {class_.class_: class_ for class_ in self.rules.last.classes}
        else:
            self.last_classes = {}
        self.property_count_by_container: dict[ContainerEntity, int] = defaultdict(int)

    def as_domain_rules(self) -> DomainRules:
        raise NotImplementedError("DomainRules not implemented yet")

    def as_asset_architect_rules(self) -> "AssetRules":
        from cognite.neat._rules.models.asset._rules import AssetClass, AssetMetadata, AssetProperty, AssetRules

        classes: SheetList[AssetClass] = SheetList[AssetClass](
            [AssetClass(**class_.model_dump()) for class_ in self.rules.classes]
        )
        properties: SheetList[AssetProperty] = SheetList[AssetProperty]()
        for prop_ in self.rules.properties:
            if prop_.type_ == EntityTypes.data_property:
                properties.append(
                    AssetProperty(**prop_.model_dump(), implementation=[AssetEntity(property=AssetFields.metadata)])
                )
            elif prop_.type_ == EntityTypes.object_property:
                properties.append(AssetProperty(**prop_.model_dump(), implementation=[RelationshipEntity()]))

        return AssetRules(
            metadata=AssetMetadata(**self.rules.metadata.model_dump()),
            properties=properties,
            classes=classes,
            prefixes=self.rules.prefixes,
        )

    def as_dms_rules(self, ignore_undefined_value_types: bool = False) -> "DMSRules":
        from cognite.neat._rules.models.dms._rules import (
            DMSContainer,
            DMSProperty,
            DMSRules,
            DMSView,
        )

        info_metadata = self.rules.metadata
        default_version = info_metadata.version
        default_space = self._to_space(info_metadata.prefix)
        metadata = self._convert_metadata_to_dms(info_metadata)

        properties_by_class: dict[ClassEntity, list[DMSProperty]] = defaultdict(list)
        referenced_containers: dict[ContainerEntity, Counter[ClassEntity]] = defaultdict(Counter)
        for prop in self.rules.properties:
            if ignore_undefined_value_types and isinstance(prop.value_type, UnknownEntity):
                continue
            dms_property = self._as_dms_property(prop, default_space, default_version)
            properties_by_class[prop.class_].append(dms_property)
            if dms_property.container:
                referenced_containers[dms_property.container][prop.class_] += 1

        views: list[DMSView] = [
            DMSView(
                class_=cls_.class_,
                name=cls_.name,
                view=cls_.class_.as_view_entity(default_space, default_version),
                description=cls_.description,
                reference=cls_.reference,
                implements=self._get_view_implements(cls_, info_metadata),
            )
            for cls_ in self.rules.classes
        ]

        last_dms_rules = _InformationRulesConverter(self.rules.last).as_dms_rules() if self.rules.last else None
        ref_dms_rules = (
            _InformationRulesConverter(self.rules.reference).as_dms_rules() if self.rules.reference else None
        )

        class_by_entity = {cls_.class_: cls_ for cls_ in self.rules.classes}
        if self.rules.last:
            for cls_ in self.rules.last.classes:
                if cls_.class_ not in class_by_entity:
                    class_by_entity[cls_.class_] = cls_

        existing_containers: set[ContainerEntity] = set()
        for rule_set in [last_dms_rules, ref_dms_rules]:
            if rule_set:
                existing_containers.update({c.container for c in rule_set.containers or []})

        containers: list[DMSContainer] = []
        for container_entity, class_entities in referenced_containers.items():
            if container_entity in existing_containers:
                continue
            constrains = self._create_container_constraint(
                class_entities, default_space, class_by_entity, referenced_containers
            )
            most_used_class_entity = class_entities.most_common(1)[0][0]
            class_ = class_by_entity[most_used_class_entity]
            container = DMSContainer(
                class_=class_.class_,
                container=container_entity,
                name=class_.name,
                description=class_.description,
                constraint=constrains or None,
            )
            containers.append(container)

        return DMSRules(
            metadata=metadata,
            properties=SheetList[DMSProperty]([prop for prop_set in properties_by_class.values() for prop in prop_set]),
            views=SheetList[DMSView](views),
            containers=SheetList[DMSContainer](containers),
            last=last_dms_rules,
            reference=ref_dms_rules,
        )

    @staticmethod
    def _create_container_constraint(
        class_entities: Counter[ClassEntity],
        default_space: str,
        class_by_entity: dict[ClassEntity, InformationClass],
        referenced_containers: Collection[ContainerEntity],
    ) -> list[ContainerEntity]:
        constrains: list[ContainerEntity] = []
        for entity in class_entities:
            class_ = class_by_entity[entity]
            for parent in class_.parent or []:
                parent_entity = parent.as_container_entity(default_space)
                if parent_entity in referenced_containers:
                    constrains.append(parent_entity)
        return constrains

    @classmethod
    def _convert_metadata_to_dms(cls, metadata: InformationMetadata) -> "DMSMetadata":
        from cognite.neat._rules.models.dms._rules import (
            DMSMetadata,
        )

        space = cls._to_space(metadata.prefix)

        return DMSMetadata(
            schema_=metadata.schema_,
            space=space,
            data_model_type=metadata.data_model_type,
            version=metadata.version,
            external_id=metadata.name.replace(" ", "_").lower(),
            creator=metadata.creator,
            name=metadata.name,
            created=metadata.created,
            updated=metadata.updated,
        )

    def _as_dms_property(self, prop: InformationProperty, default_space: str, default_version: str) -> "DMSProperty":
        """This creates the first"""

        from cognite.neat._rules.models.dms._rules import DMSProperty

        # returns property type, which can be ObjectProperty or DatatypeProperty
        value_type: DataType | ViewEntity | DMSUnknownEntity
        if isinstance(prop.value_type, DataType):
            value_type = prop.value_type

        # UnknownEntity should  resolve to DMSUnknownEntity
        # meaning end node type is unknown
        elif isinstance(prop.value_type, UnknownEntity):
            value_type = DMSUnknownEntity()

        elif isinstance(prop.value_type, ClassEntity):
            value_type = prop.value_type.as_view_entity(default_space, default_version)

        elif isinstance(prop.value_type, MultiValueTypeInfo):
            # Multi Object type should resolve to DMSUnknownEntity
            # meaning end node type is unknown
            if prop.value_type.is_multi_object_type():
                value_type = DMSUnknownEntity()

            # Multi Data type should resolve to a single data type, or it should
            elif prop.value_type.is_multi_data_type():
                value_type = self.convert_multi_data_type(prop.value_type)

            # Mixed types default to string
            else:
                value_type = String()

        else:
            raise ValueError(f"Unsupported value type: {prop.value_type.type_}")

        connection: Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None = None
        if isinstance(value_type, ViewEntity):
            # Default connection type.
            connection = EdgeEntity() if prop.is_list else "direct"

        # defaulting to direct connection
        elif isinstance(value_type, DMSUnknownEntity):
            connection = "direct"

        container: ContainerEntity | None = None
        container_property: str | None = None
        is_list: bool | None = prop.is_list
        nullable: bool | None = not prop.is_mandatory
        if isinstance(connection, EdgeEntity):
            nullable = None
        elif connection == "direct":
            nullable = True
            container, container_property = self._get_container(prop, default_space)
        else:
            container, container_property = self._get_container(prop, default_space)

        return DMSProperty(
            class_=prop.class_,
            name=prop.name,
            property_=prop.property_,
            value_type=value_type,
            nullable=nullable,
            is_list=is_list,
            connection=connection,
            default=prop.default,
            reference=prop.reference,
            container=container,
            container_property=container_property,
            view=prop.class_.as_view_entity(default_space, default_version),
            view_property=prop.property_,
        )

    @classmethod
    def _to_space(cls, prefix: str) -> str:
        """Ensures that the prefix comply with the CDF space regex"""
        prefix = re.sub(r"[^a-zA-Z0-9_-]", "_", prefix)
        if prefix[0].isdigit() or prefix[0] == "_":
            prefix = f"a{prefix}"
        prefix = prefix[:43]
        if prefix[-1] == "_":
            prefix = f"{prefix[:-1]}1"
        return prefix

    def _get_container(self, prop: InformationProperty, default_space: str) -> tuple[ContainerEntity, str]:
        if isinstance(prop.reference, ReferenceEntity):
            return (
                prop.reference.as_container_entity(default_space),
                prop.reference.property_ or prop.property_,
            )
        elif (self.is_addition or self.is_reshape) and prop.class_ in self.last_classes:
            # We need to create a new container for the property, as we cannot change
            # the existing container in the last schema
            container_entity = prop.class_.as_container_entity(default_space)
            container_entity.suffix = self._bump_suffix(container_entity.suffix)
        else:
            container_entity = prop.class_.as_container_entity(default_space)

        while self.property_count_by_container[container_entity] >= DMS_CONTAINER_PROPERTY_SIZE_LIMIT:
            container_entity.suffix = self._bump_suffix(container_entity.suffix)

        self.property_count_by_container[container_entity] += 1
        return container_entity, prop.property_

    def _get_view_implements(self, cls_: InformationClass, metadata: InformationMetadata) -> list[ViewEntity]:
        if isinstance(cls_.reference, ReferenceEntity) and cls_.reference.prefix != metadata.prefix:
            # We use the reference for implements if it is in a different namespace
            if self.rules.reference and cls_.reference.prefix == self.rules.reference.metadata.prefix:
                implements = [
                    cls_.reference.as_view_entity(
                        self.rules.reference.metadata.prefix, self.rules.reference.metadata.version
                    )
                ]
            else:
                implements = [
                    cls_.reference.as_view_entity(metadata.prefix, metadata.version),
                ]
        else:
            implements = []
        for parent in cls_.parent or []:
            if self.rules.reference and parent.prefix == self.rules.reference.metadata.prefix:
                view_entity = parent.as_view_entity(
                    self.rules.reference.metadata.prefix, self.rules.reference.metadata.version
                )
            else:
                view_entity = parent.as_view_entity(metadata.prefix, metadata.version)
            implements.append(view_entity)
        return implements

    @staticmethod
    def _bump_suffix(suffix: str) -> str:
        suffix_number = re.search(r"\d+$", suffix)

        if suffix_number:
            return suffix[: suffix_number.start()] + str(int(suffix_number.group()) + 1)
        else:
            return f"{suffix}2"

    @staticmethod
    def convert_multi_data_type(value_type: MultiValueTypeInfo) -> DataType:
        if not value_type.is_multi_data_type():
            raise ValueError("Only MultiValueType with DataType types is supported")
        # We check above that there are no ClassEntity types in the MultiValueType
        py_types = {type_.python for type_ in value_type.types}  # type: ignore[union-attr]

        # JSON mixed with other types should resolve to string that is safe choice
        if dms.Json in py_types and len(py_types) > 1:
            return data_types.String()
        elif dms.Json in py_types:
            return data_types.Json()
        elif not (py_types - {bool}):
            return data_types.Boolean()
        elif not (py_types - {int, bool}):
            return data_types.Integer()
        elif not (py_types - {float, int, bool}):
            return data_types.Double()
        elif not (py_types - {datetime, date}):
            return data_types.DateTime()

        return data_types.String()


class _DMSRulesConverter:
    def __init__(self, dms: DMSRules):
        self.dms = dms

    def as_domain_rules(self) -> "DomainRules":
        raise NotImplementedError("DomainRules not implemented yet")

    def as_information_rules(
        self,
    ) -> "InformationRules":
        from cognite.neat._rules.models.information._rules import (
            InformationClass,
            InformationProperty,
            InformationRules,
        )

        dms = self.dms.metadata

        metadata = self._convert_metadata_to_info(dms)

        classes = [
            InformationClass(
                # we do not want a version in class as we use URI for the class
                class_=ClassEntity(prefix=view.class_.prefix, suffix=view.class_.suffix),
                description=view.description,
                parent=[
                    # we do not want a version in class as we use URI for the class
                    implemented_view.as_class(skip_version=True)
                    # We only want parents in the same namespace, parent in a different namespace is a reference
                    for implemented_view in view.implements or []
                    if implemented_view.prefix == view.class_.prefix
                ],
                reference=self._get_class_reference(view),
            )
            for view in self.dms.views
        ]

        properties: list[InformationProperty] = []
        value_type: DataType | ClassEntity | str
        for property_ in self.dms.properties:
            if isinstance(property_.value_type, DataType):
                value_type = property_.value_type
            elif isinstance(property_.value_type, ViewEntity):
                value_type = ClassEntity(
                    prefix=property_.value_type.prefix,
                    suffix=property_.value_type.suffix,
                )
            elif isinstance(property_.value_type, DMSUnknownEntity):
                value_type = UnknownEntity()
            else:
                raise ValueError(f"Unsupported value type: {property_.value_type.type_}")

            properties.append(
                InformationProperty(
                    # Removing version
                    class_=ClassEntity(suffix=property_.class_.suffix, prefix=property_.class_.prefix),
                    property_=property_.view_property,
                    value_type=value_type,
                    description=property_.description,
                    min_count=0 if property_.nullable or property_.nullable is None else 1,
                    max_count=float("inf") if property_.is_list or property_.nullable is None else 1,
                    reference=self._get_property_reference(property_),
                )
            )

        return InformationRules(
            metadata=metadata,
            properties=SheetList[InformationProperty](properties),
            classes=SheetList[InformationClass](classes),
            last=_DMSRulesConverter(self.dms.last).as_information_rules() if self.dms.last else None,
            reference=_DMSRulesConverter(self.dms.reference).as_information_rules() if self.dms.reference else None,
        )

    @classmethod
    def _convert_metadata_to_info(cls, metadata: DMSMetadata) -> "InformationMetadata":
        from cognite.neat._rules.models.information._rules import InformationMetadata

        prefix = metadata.space
        return InformationMetadata(
            schema_=metadata.schema_,
            data_model_type=metadata.data_model_type,
            extension=metadata.extension,
            prefix=prefix,
            namespace=Namespace(f"https://purl.orgl/neat/{prefix}/"),
            version=metadata.version,
            description=metadata.description,
            name=metadata.name or metadata.external_id,
            creator=metadata.creator,
            created=metadata.created,
            updated=metadata.updated,
        )

    @classmethod
    def _get_class_reference(cls, view: DMSView) -> ReferenceEntity | None:
        parents_other_namespace = [parent for parent in view.implements or [] if parent.prefix != view.class_.prefix]
        if len(parents_other_namespace) == 0:
            return None
        if len(parents_other_namespace) > 1:
            warnings.warn(
                ParentInDifferentSpaceWarning(view.view.as_id()),
                stacklevel=2,
            )
        other_parent = parents_other_namespace[0]

        return ReferenceEntity(prefix=other_parent.prefix, suffix=other_parent.suffix)

    @classmethod
    def _get_property_reference(cls, property_: DMSProperty) -> ReferenceEntity | None:
        has_container_other_namespace = property_.container and property_.container.prefix != property_.class_.prefix
        if not has_container_other_namespace:
            return None
        container = cast(ContainerEntity, property_.container)
        return ReferenceEntity(
            prefix=container.prefix,
            suffix=container.suffix,
            property=property_.container_property,
        )



from abc import ABC, abstractmethod
from collections.abc import Iterable
from pathlib import Path
from typing import Generic, TypeVar

from cognite.client import CogniteClient

from cognite.neat._rules._shared import T_VerifiedRules
from cognite.neat._utils.auxiliary import class_html_doc
from cognite.neat._utils.upload import UploadResult, UploadResultList

T_Export = TypeVar("T_Export")


class BaseExporter(ABC, Generic[T_VerifiedRules, T_Export]):
    _new_line = "\n"
    _encoding = "utf-8"

    @abstractmethod
    def export_to_file(self, rules: T_VerifiedRules, filepath: Path) -> None:
        raise NotImplementedError

    @abstractmethod
    def export(self, rules: T_VerifiedRules) -> T_Export:
        raise NotImplementedError

    @classmethod
    def _repr_html_(cls) -> str:
        return class_html_doc(cls, include_factory_methods=False)


class CDFExporter(BaseExporter[T_VerifiedRules, T_Export]):
    @abstractmethod
    def export_to_cdf_iterable(
        self, rules: T_VerifiedRules, client: CogniteClient, dry_run: bool = False, fallback_one_by_one: bool = False
    ) -> Iterable[UploadResult]:
        raise NotImplementedError

    def export_to_cdf(
        self, rules: T_VerifiedRules, client: CogniteClient, dry_run: bool = False, fallback_one_by_one: bool = False
    ) -> UploadResultList:
        return UploadResultList(self.export_to_cdf_iterable(rules, client, dry_run, fallback_one_by_one))



from ._rules import AssetClass, AssetMetadata, AssetProperty, AssetRules
from ._rules_input import AssetInputClass, AssetInputMetadata, AssetInputProperty, AssetInputRules

__all__ = [
    "AssetRules",
    "AssetMetadata",
    "AssetClass",
    "AssetProperty",
    "AssetInputRules",
    "AssetInputMetadata",
    "AssetInputClass",
    "AssetInputProperty",
]



import datetime

import numpy as np
import pandas as pd
from rdflib import OWL, Literal, Namespace

from cognite.neat._rules._constants import PATTERNS
from cognite.neat._rules.models._base_rules import MatchType
from cognite.neat._rules.models.data_types import _XSD_TYPES
from cognite.neat._utils.rdf_ import remove_namespace_from_uri


def parse_raw_classes_dataframe(query_results: list[tuple]) -> pd.DataFrame:
    df = pd.DataFrame(
        query_results,
        columns=[
            "Class",
            "Name",
            "Description",
            "Parent Class",
            "Reference",
            "Match",
            "Comment",
        ],
    )

    if df.empty:
        return df

    # # remove NaNs
    df.replace(np.nan, "", regex=True, inplace=True)

    df.Class = df.Class.apply(lambda x: remove_namespace_from_uri(x))
    df["Match Type"] = len(df) * [MatchType.exact]
    df["Comment"] = len(df) * [None]
    df["Parent Class"] = df["Parent Class"].apply(lambda x: remove_namespace_from_uri(x))

    return df


def clean_up_classes(df: pd.DataFrame) -> pd.DataFrame:
    clean_list = [
        {
            "Class": class_,
            "Name": group_df["Name"].unique()[0],
            "Description": "\n".join(list(group_df.Description.unique())),
            "Parent Class": ", ".join(list(group_df["Parent Class"].unique())),
            "Reference": group_df["Reference"].unique()[0],
            "Match Type": group_df["Match Type"].unique()[0],
            "Comment": group_df["Comment"].unique()[0],
        }
        for class_, group_df in df.groupby("Class")
    ]

    df = pd.DataFrame(clean_list)

    # bring NaNs back
    df.replace("", None, inplace=True)

    # split Parent Class column back into list
    df["Parent Class"] = df["Parent Class"].apply(lambda x: x.split(", ") if isinstance(x, str) else None)

    return df


def make_classes_compliant(classes: pd.DataFrame, importer: str = "RDF-based") -> pd.DataFrame:
    """Make classes compliant.

    Returns:
        Dataframe containing compliant classes

    !!! note "About the compliant classes"
        The compliant classes are based on the OWL base ontology, but adapted to NEAT and use in CDF.
        One thing to note is that this method would not be able to fix issues with class ids which
        are not compliant with the CDF naming convention. For example, if a class id contains a space,
        starts with a number, etc. This will cause issues when trying to create the class in CDF.
    """

    # Replace empty or non-string values in "Match" column with "exact"
    classes["Match Type"] = classes["Match Type"].fillna(MatchType.exact)
    classes["Match Type"] = classes["Match Type"].apply(
        lambda x: MatchType.exact if not isinstance(x, str) or len(x) == 0 else x
    )

    # Replace empty or non-string values in "Comment" column with a default value
    classes["Comment"] = classes["Comment"].fillna(f"Imported using {importer} importer")
    classes["Comment"] = classes["Comment"].apply(
        lambda x: (f"Imported using {importer} importer" if not isinstance(x, str) or len(x) == 0 else x)
    )

    # Add _object_property_class, _data_type_property_class, _thing_class to the dataframe
    classes = pd.concat(
        [
            classes,
            pd.DataFrame([object_property_class(), data_type_property_class(), thing_class()]),
        ],
        ignore_index=True,
    )

    # Reduce length of elements in the "Description" column to 1024 characters
    classes["Description"] = classes["Description"].apply(lambda x: x[:1024] if isinstance(x, str) else None)

    # Add missing parent classes to the dataframe
    classes = pd.concat(
        [classes, pd.DataFrame(add_parent_class(classes))],
        ignore_index=True,
    )

    return classes


def object_property_class() -> dict:
    return {
        "Class": "ObjectProperty",
        "Name": None,
        "Description": "The class of object properties.",
        "Parent Class": None,
        "Reference": OWL.ObjectProperty,
        "Match Type": MatchType.exact,
        "Comment": "Added by NEAT based on owl:ObjectProperty but adapted to NEAT and use in CDF.",
    }


def data_type_property_class() -> dict:
    return {
        "Class": "DatatypeProperty",
        "Name": None,
        "Description": "The class of data properties.",
        "Parent Class": None,
        "Reference": OWL.DatatypeProperty,
        "Match Type": MatchType.exact,
        "Comment": "Added by NEAT based on owl:DatatypeProperty but adapted to NEAT and use in CDF.",
    }


def thing_class() -> dict:
    return {
        "Class": "Thing",
        "Name": None,
        "Description": "The class of holding class individuals.",
        "Parent Class": None,
        "Reference": OWL.Thing,
        "Match Type": MatchType.exact,
        "Comment": (
            "Added by NEAT. "
            "Imported from OWL base ontology, it is meant for use as a default"
            " value type for object properties which miss a declared range."
        ),
    }


def add_parent_class(df: pd.DataFrame) -> list[dict]:
    parent_set = {
        item
        for sublist in df["Parent Class"].tolist()
        if sublist
        for item in sublist
        if item != "" and item is not None
    }
    class_set = set(df["Class"].tolist())

    rows = []
    for missing_parent_class in parent_set.difference(class_set):
        rows += [
            {
                "Class": missing_parent_class,
                "Name": None,
                "Description": None,
                "Parent Class": None,
                "Reference": None,
                "Match Type": None,
                "Comment": (
                    "Added by NEAT. "
                    "This is a parent class that is missing in the ontology. "
                    "It is added by NEAT to make the ontology compliant with CDF."
                ),
            }
        ]

    return rows


def parse_raw_properties_dataframe(query_results: list[tuple]) -> pd.DataFrame:
    df = pd.DataFrame(
        query_results,
        columns=[
            "Class",
            "Property",
            "Name",
            "Description",
            "Value Type",
            "Min Count",
            "Max Count",
            "Default",
            "Reference",
            "Match Type",
            "Comment",
            "_property_type",
        ],
    )
    if df.empty:
        return df

    df.replace(np.nan, "", regex=True, inplace=True)

    df.Class = df.Class.apply(lambda x: remove_namespace_from_uri(x))
    df.Property = df.Property.apply(lambda x: remove_namespace_from_uri(x))
    df["Value Type"] = df["Value Type"].apply(lambda x: remove_namespace_from_uri(x))
    df["Match Type"] = len(df) * [MatchType.exact]
    df["Comment"] = len(df) * [None]
    df["_property_type"] = df["_property_type"].apply(lambda x: remove_namespace_from_uri(x))

    return df


def clean_up_properties(df: pd.DataFrame) -> pd.DataFrame:
    class_grouped_dfs = df.groupby("Class")

    clean_list = []

    for class_, class_grouped_df in class_grouped_dfs:
        property_grouped_dfs = class_grouped_df.groupby("Property")
        for property_, property_grouped_df in property_grouped_dfs:
            clean_list += [
                {
                    "Class": class_,
                    "Property": property_,
                    "Name": property_grouped_df["Name"].unique()[0],
                    "Description": "\n".join(list(property_grouped_df["Description"].unique()))[:1024],
                    "Value Type": property_grouped_df["Value Type"].unique()[0],
                    "Min Count": property_grouped_df["Min Count"].unique()[0],
                    "Max Count": property_grouped_df["Max Count"].unique()[0],
                    "Default": property_grouped_df["Default"].unique()[0],
                    "Reference": property_grouped_df["Reference"].unique()[0],
                    "Match Type": property_grouped_df["Match Type"].unique()[0],
                    "Comment": property_grouped_df["Comment"].unique()[0],
                    "_property_type": property_grouped_df["_property_type"].unique()[0],
                }
            ]

    df = pd.DataFrame(clean_list)
    df.replace("", None, inplace=True)

    return df


def make_properties_compliant(properties: pd.DataFrame, importer: str = "RDF-based") -> pd.DataFrame:
    # default to 0 if "Min Count" is not specified
    properties["Min Count"] = properties["Min Count"].apply(lambda x: 0 if not isinstance(x, Literal) or x == "" else x)

    # default to 1 if "Max Count" is not specified
    properties["Max Count"] = properties["Max Count"].apply(lambda x: 1 if not isinstance(x, Literal) or x == "" else x)

    # Replace empty or non-string values in "Match Type" column with "exact"
    properties["Match Type"] = properties["Match Type"].fillna("exact")
    properties["Match Type"] = properties["Match Type"].apply(
        lambda x: "exact" if not isinstance(x, str) or len(x) == 0 else x
    )

    # Replace empty or non-string values in "Comment" column with a default value
    properties["Comment"] = properties["Comment"].fillna(f"Imported using {importer} importer")
    properties["Comment"] = properties["Comment"].apply(
        lambda x: (f"Imported using {importer} importer" if not isinstance(x, str) or len(x) == 0 else x)
    )

    # Reduce length of elements in the "Description" column to 1024 characters
    properties["Description"] = properties["Description"].apply(lambda x: x[:1024] if isinstance(x, str) else None)

    # fixes and additions
    properties = fix_dangling_properties(properties)
    properties = fix_missing_property_value_type(properties)

    return properties


def fix_dangling_properties(properties: pd.DataFrame) -> pd.DataFrame:
    """This method fixes properties which are missing a domain definition in the ontology.

    Args:
        properties: Dataframe containing properties

    Returns:
        Dataframe containing properties with fixed domain
    """
    domain = {
        "ObjectProperty": object_property_class()["Class"],
        "DatatypeProperty": data_type_property_class()["Class"],
    }

    # apply missing range
    properties["Class"] = properties.apply(
        lambda row: (
            domain[row._property_type]
            if row._property_type == "ObjectProperty" and pd.isna(row["Class"])
            else domain["DatatypeProperty"]
            if pd.isna(row["Class"])
            else row["Class"]
        ),
        axis=1,
    )
    return properties


def fix_missing_property_value_type(properties: pd.DataFrame) -> pd.DataFrame:
    """This method fixes properties which are missing a range definition in the ontology.

    Args:
        properties: Dataframe containing properties

    Returns:
        Dataframe containing properties with fixed range
    """
    # apply missing range
    properties["Value Type"] = properties.apply(
        lambda row: (
            thing_class()["Class"]
            if row._property_type == "ObjectProperty" and pd.isna(row["Value Type"])
            else "string"
            if pd.isna(row["Value Type"])
            else row["Value Type"]
        ),
        axis=1,
    )

    return properties


def make_metadata_compliant(metadata: dict) -> dict:
    """Attempts to fix errors in metadata, otherwise defaults to values that will pass validation.

    Args:
        metadata: Dictionary containing metadata

    Returns:
        Dictionary containing metadata with fixed errors
    """

    metadata = fix_namespace(metadata, default=Namespace("http://purl.org/cognite/neat#"))
    metadata = fix_prefix(metadata)
    metadata = fix_version(metadata)
    metadata = fix_date(
        metadata,
        date_type="created",
        default=datetime.datetime.now().replace(microsecond=0),
    )
    metadata = fix_date(
        metadata,
        date_type="updated",
        default=datetime.datetime.now().replace(microsecond=0),
    )
    metadata = fix_title(metadata)
    metadata = fix_description(metadata)
    metadata = fix_author(metadata, "creator")
    metadata = fix_rights(metadata)
    metadata = fix_license(metadata)

    return metadata


def fix_license(metadata: dict, default: str = "Unknown license") -> dict:
    if license := metadata.get("license", None):
        if not isinstance(license, str):
            metadata["license"] = default
        elif isinstance(license, str) and len(license) == 0:
            metadata["license"] = default
    else:
        metadata["license"] = default
    return metadata


def fix_rights(metadata: dict, default: str = "Unknown rights") -> dict:
    if rights := metadata.get("rights", None):
        if not isinstance(rights, str):
            metadata["rights"] = default
        elif isinstance(rights, str) and len(rights) == 0:
            metadata["rights"] = default
    else:
        metadata["rights"] = default
    return metadata


def fix_author(metadata: dict, author_type: str = "creator", default: str = "NEAT") -> dict:
    if author := metadata.get(author_type, None):
        if not isinstance(author, str) or isinstance(author, list):
            metadata[author_type] = default
        elif isinstance(author, str) and len(author) == 0:
            metadata[author_type] = default
    else:
        metadata[author_type] = default
    return metadata


def fix_description(metadata: dict, default: str = "This model has been inferred from OWL ontology") -> dict:
    if description := metadata.get("description", None):
        if not isinstance(description, str) or len(description) == 0:
            metadata["description"] = default
        elif isinstance(description, str) and len(description) > 1024:
            metadata["description"] = metadata["description"][:1024]
    else:
        metadata["description"] = default
    return metadata


def fix_prefix(metadata: dict, default: str = "neat") -> dict:
    if prefix := metadata.get("prefix", None):
        if not isinstance(prefix, str) or not PATTERNS.prefix_compliance.match(prefix):
            metadata["prefix"] = default
    else:
        metadata["prefix"] = default
    return metadata


def fix_namespace(metadata: dict, default: Namespace) -> dict:
    if namespace := metadata.get("namespace", None):
        if not isinstance(namespace, Namespace):
            try:
                metadata["namespace"] = Namespace(namespace)
            except Exception:
                metadata["namespace"] = default
    else:
        metadata["namespace"] = default

    return metadata


def fix_date(
    metadata: dict,
    date_type: str,
    default: datetime.datetime,
) -> dict:
    if date := metadata.get(date_type, None):
        try:
            if isinstance(date, datetime.datetime):
                return metadata
            elif isinstance(date, datetime.date):
                metadata[date_type] = datetime.datetime.combine(metadata[date_type], datetime.datetime.min.time())
            elif isinstance(date, str):
                metadata[date_type] = datetime.datetime.strptime(metadata[date_type], "%Y-%m-%dT%H:%M:%SZ")
            else:
                metadata[date_type] = default
        except Exception:
            metadata[date_type] = default
    else:
        metadata[date_type] = default

    return metadata


def fix_version(metadata: dict, default: str = "1.0.0") -> dict:
    if version := metadata.get("version", None):
        if not PATTERNS.version_compliance.match(version):
            metadata["version"] = default
    else:
        metadata["version"] = default

    return metadata


def fix_title(metadata: dict, default: str = "OWL Inferred Data Model") -> dict:
    if title := metadata.get("title", None):
        if not isinstance(title, str):
            metadata["title"] = default
        elif isinstance(title, str) and len(title) == 0:
            metadata["title"] = default
        elif isinstance(title, str) and len(title) > 255:
            metadata["title"] = metadata["title"][:255]
        else:
            pass
    else:
        metadata["title"] = default

    return metadata


def make_components_compliant(components: dict) -> dict:
    components = add_missing_classes(components)
    components = add_missing_value_types(components)
    components = add_default_property_to_dangling_classes(components)

    return components


def add_missing_classes(components: dict[str, list[dict]]) -> dict:
    """Add missing classes to Classes.

    Args:
        tables: imported tables from owl ontology

    Returns:
        Updated tables with missing classes added to containers
    """

    missing_classes = {definition["Class"] for definition in components["Properties"]} - {
        definition["Class"] for definition in components["Classes"]
    }

    comment = (
        "Added by NEAT. "
        "This is a class that a domain of a property but was not defined in the ontology. "
        "It is added by NEAT to make the ontology compliant with CDF."
    )

    for class_ in missing_classes:
        components["Classes"].append({"Class": class_, "Comment": comment})

    return components


def add_missing_value_types(components: dict) -> dict:
    """Add properties to classes that do not have any properties defined to them

    Args:
        tables: imported tables from owl ontology

    Returns:
        Updated tables with missing properties added to containers
    """

    xsd_types = _XSD_TYPES
    candidate_value_types = {definition["Value Type"] for definition in components["Properties"]} - {
        definition["Class"] for definition in components["Classes"]
    }

    # to avoid issue of case sensitivity for xsd types
    value_types_lower = {v.lower() for v in candidate_value_types}

    xsd_types_lower = {x.lower() for x in xsd_types}

    # Create a mapping from lowercase strings to original strings
    value_types_mapping = {v.lower(): v for v in candidate_value_types}

    # Find the difference
    difference = value_types_lower - xsd_types_lower

    # Convert the difference back to the original case
    difference_original_case = {value_types_mapping[d] for d in difference}

    for class_ in difference_original_case:
        components["Classes"].append(
            {
                "Class": class_,
                "Comment": (
                    "Added by NEAT. "
                    "This is a class that a domain of a property but was not defined in the ontology. "
                    "It is added by NEAT to make the ontology compliant with CDF."
                ),
            }
        )

    return components


def add_default_property_to_dangling_classes(components: dict[str, list[dict]]) -> dict:
    """Add missing classes to Classes.

    Args:
        tables: imported tables from owl ontology

    Returns:
        Updated tables with missing classes added to containers
    """

    dangling_classes = {
        definition["Class"] for definition in components["Classes"] if not definition.get("Parent Class", None)
    } - {definition["Class"] for definition in components["Properties"]}

    comment = (
        "Added by NEAT. "
        "This is property has been added to this class since otherwise it will create "
        "dangling classes in the ontology."
    )

    for class_ in dangling_classes:
        components["Properties"].append(
            {
                "Class": class_,
                "Property": "label",
                "Value Type": "string",
                "Comment": comment,
                "Min Count": 0,
                "Max Count": 1,
                "Reference": "http://www.w3.org/2000/01/rdf-schema#label",
            }
        )

    return components



from typing import ClassVar

from cognite.neat._store import NeatGraphStore
from cognite.neat._workflows.model import FlowMessage
from cognite.neat._workflows.steps.data_contracts import (
    MultiRuleData,
    NeatGraph,
)
from cognite.neat._workflows.steps.step_model import Configurable, Step

__all__ = ["GraphStoreConfiguration"]

CATEGORY = __name__.split(".")[-1].replace("_", " ").title()


class GraphStoreConfiguration(Step):
    """
    This step initializes source OR solution graph store
    """

    description = "This step initializes the source and solution graph stores."
    version = "private-beta"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="Graph store type",
            value="oxigraph",
            label="Graph store type, supported: oxigraph, rdflib",
            options=["oxigraph", "rdflib"],
        ),
    ]

    def run(  # type: ignore[override]
        self, rules_data: MultiRuleData | None = None
    ) -> (FlowMessage, NeatGraph):  # type: ignore[syntax]
        store_type = self.configs.get("Graph store type", "oxigraph")

        if store_type == "oxigraph":
            store = NeatGraph(graph=NeatGraphStore.from_oxi_store(rules=rules_data.information if rules_data else None))
        else:
            store = NeatGraph(
                graph=NeatGraphStore.from_memory_store(rules=rules_data.information if rules_data else None)
            )

        return (
            FlowMessage(output_text="Graph store configured successfully"),
            store,
        )



"""This module performs importing of various formats to one of serializations for which
there are loaders to TransformationRules pydantic class."""

from cognite.neat._rules.importers._rdf._base import BaseRDFImporter
from cognite.neat._rules.importers._rdf._shared import make_components_compliant

from ._owl2classes import parse_owl_classes
from ._owl2metadata import parse_owl_metadata
from ._owl2properties import parse_owl_properties


class OWLImporter(BaseRDFImporter):
    """Convert OWL ontology to tables/ transformation rules / Excel file.

        Args:
            filepath: Path to OWL ontology

    !!! Note
        OWL Ontologies are information models which completeness varies. As such, constructing functional
        data model directly will often be impossible, therefore the produced Rules object will be ill formed.
        To avoid this, neat will automatically attempt to make the imported rules compliant by adding default
        values for missing information, attaching dangling properties to default containers based on the
        property type, etc.

        One has to be aware that NEAT will be opinionated about how to make the ontology
        compliant, and that the resulting rules may not be what you expect.

    """

    def _to_rules_components(
        self,
    ) -> dict:
        components = {
            "Metadata": parse_owl_metadata(self.graph),
            "Classes": parse_owl_classes(self.graph),
            "Properties": parse_owl_properties(self.graph),
        }

        return make_components_compliant(components)



"""This module contains the definition of `TransformationRules` pydantic model and all
its sub-models and validators.
"""

from __future__ import annotations

import sys
import types
from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable, Iterator, MutableSequence, Sequence
from typing import Annotated, Any, ClassVar, Literal, SupportsIndex, TypeVar, get_args, get_origin, overload

import pandas as pd
from pydantic import (
    BaseModel,
    BeforeValidator,
    ConfigDict,
    Field,
    GetCoreSchemaHandler,
    PlainSerializer,
    field_validator,
    model_serializer,
)
from pydantic.main import IncEx
from pydantic_core import core_schema

from cognite.neat._rules.models._types import ClassEntityType, InformationPropertyType

if sys.version_info >= (3, 11):
    from enum import StrEnum
    from typing import Self
else:
    from backports.strenum import StrEnum
    from typing_extensions import Self


METADATA_VALUE_MAX_LENGTH = 5120


def _get_required_fields(model: type[BaseModel], use_alias: bool = False) -> set[str]:
    """Get required fields from a pydantic model.

    Parameters
    ----------
    model : type[BaseModel]
        Pydantic data model
    use_alias : bool, optional
        Whether to return field alias name, by default False

    Returns
    -------
    list[str]
        List of required fields
    """
    required_fields = set()
    for name, field in model.model_fields.items():
        if not field.is_required():
            continue

        alias = getattr(field, "alias", None)
        if use_alias and alias:
            required_fields.add(alias)
        else:
            required_fields.add(name)
    return required_fields


class SchemaCompleteness(StrEnum):
    complete = "complete"
    partial = "partial"
    extended = "extended"


class ExtensionCategory(StrEnum):
    addition = "addition"
    reshape = "reshape"
    rebuild = "rebuild"


class DataModelType(StrEnum):
    solution = "solution"
    enterprise = "enterprise"


class RoleTypes(StrEnum):
    domain_expert = "domain expert"
    information = "information architect"
    asset = "asset architect"
    dms = "DMS Architect"


class MatchType(StrEnum):
    exact = "exact"
    partial = "partial"


class SchemaModel(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(
        populate_by_name=True,
        str_strip_whitespace=True,
        arbitrary_types_allowed=True,
        strict=False,
        extra="ignore",
        use_enum_values=True,
    )
    validators_to_skip: set[str] = Field(default_factory=set, exclude=True)

    @classmethod
    def mandatory_fields(cls, use_alias=False) -> set[str]:
        """Returns a set of mandatory fields for the model."""
        return _get_required_fields(cls, use_alias)

    @field_validator("*", mode="before")
    def strip_string(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.strip()
        elif isinstance(value, list):
            return [entry.strip() if isinstance(entry, str) else entry for entry in value]
        return value


class BaseMetadata(SchemaModel):
    """
    Metadata model for data model
    """

    role: ClassVar[RoleTypes]

    def to_pandas(self) -> pd.Series:
        """Converts Metadata to pandas Series."""
        return pd.Series(self.model_dump())

    def _repr_html_(self) -> str:
        """Returns HTML representation of Metadata."""
        return self.to_pandas().to_frame("value")._repr_html_()  # type: ignore[operator]

    @classmethod
    def mandatory_fields(cls, use_alias=False) -> set[str]:
        """Returns a set of mandatory fields for the model."""
        return _get_required_fields(cls, use_alias)

    @model_serializer(mode="wrap")
    def include_role(self, serializer: Callable) -> dict:
        return {"role": self.role.value, **serializer(self)}

    @abstractmethod
    def as_identifier(self) -> str:
        """Returns a unique identifier for the metadata."""
        raise NotImplementedError()

    @abstractmethod
    def get_prefix(self) -> str:
        """Returns the prefix for the metadata."""
        raise NotImplementedError()


class BaseRules(SchemaModel, ABC):
    """
    Rules is a core concept in `neat`. This represents fusion of data model
    definitions and (optionally) the transformation rules used to transform the data/graph
    from the source representation to the target representation defined by the data model.
    The rules are defined in an Excel sheet and then parsed into a `Rules` object. The
    `Rules` object is then used to generate data model and the `RDF` graph made of data
    model instances.

    Args:
        metadata: Data model metadata
        validators_to_skip: List of validators to skip. Defaults to []
    """

    metadata: BaseMetadata
    reference: Self | None = Field(None, alias="Reference")

    @classmethod
    def headers_by_sheet(cls, by_alias: bool = False) -> dict[str, list[str]]:
        """Returns a list of headers for the model, typically used by ExcelExporter"""
        headers_by_sheet: dict[str, list[str]] = {}
        for field_name, field in cls.model_fields.items():
            if field_name == "validators_to_skip":
                continue
            sheet_name = (field.alias or field_name) if by_alias else field_name
            annotation = field.annotation

            if isinstance(annotation, types.UnionType):
                annotation = annotation.__args__[0]

            try:
                if isinstance(annotation, types.GenericAlias) and get_origin(annotation) is SheetList:
                    # We know that this is a SheetList, so we can safely access the annotation
                    # which is the concrete type of the SheetEntity.
                    model_fields = get_args(annotation)[0].model_fields  # type: ignore[union-attr]
                elif isinstance(annotation, type) and issubclass(annotation, BaseModel):
                    model_fields = annotation.model_fields
                else:
                    model_fields = {}
            except TypeError:
                # Python 3.10 raises TypeError: issubclass() arg 1 must be a class
                # when calling issubclass(annotation, SheetList) with the dict annotation
                model_fields = {}
            headers_by_sheet[sheet_name] = [
                (field.alias or field_name) if by_alias else field_name
                for field_name, field in model_fields.items()
                if field_name != "validators_to_skip" and not field.exclude
            ]
        return headers_by_sheet

    def dump(
        self,
        entities_exclude_defaults: bool = True,
        as_reference: bool = False,
        mode: Literal["python", "json"] = "python",
        by_alias: bool = False,
        exclude: IncEx | None = None,
        exclude_none: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
    ) -> dict[str, Any]:
        """Dump the model to a dictionary.

        This is used in the Exporters to dump rules in the required format.

        Args:
            entities_exclude_defaults: Whether to exclude default prefix (and version) for entities.
                For example, given a class that is dumped as 'my_prefix:MyClass', if the prefix for the rules
                set in metadata.prefix = 'my_prefix', then this class will be dumped as 'MyClass' when this flag is set.
                Defaults to True.
            as_reference (bool, optional): Whether to dump as reference. For Information and DMS rules, this will
                set the reference column/field to the reference of that entity. This is used in the ExcelExporter
                to dump a reference model.
            mode: The mode in which `to_python` should run.
                If mode is 'json', the output will only contain JSON serializable types.
                If mode is 'python', the output may contain non-JSON-serializable Python objects.
            by_alias: Whether to use the field's alias in the dictionary key if defined.
            exclude: A set of fields to exclude from the output.
            exclude_none: Whether to exclude fields that have a value of `None`.
            exclude_unset: Whether to exclude fields that have not been explicitly set.
            exclude_defaults: Whether to exclude fields that are set to their default value.
        """
        for field_name in self.model_fields.keys():
            value = getattr(self, field_name)
            # Ensure deterministic order of properties, classes, views, and so on
            if isinstance(value, SheetList):
                value.sort(key=lambda x: x._identifier())

        context: dict[str, Any] = {"as_reference": as_reference}
        if entities_exclude_defaults:
            context["metadata"] = self.metadata

        exclude_input: IncEx | None
        if self.reference is None:
            exclude_input = exclude
        else:
            # If the rules has a reference, we dump that separately with the as_reference flag set to True.
            # We don't want to include the reference in the main dump, so we exclude it here.
            # This is to include whatever is in the exclude set from the user.
            if isinstance(exclude, dict):
                exclude_input = exclude.copy()
                exclude_input["reference"] = {"__all__"}  # type: ignore[index]
            elif isinstance(exclude, set):
                exclude_input = exclude.copy()
                exclude_input.add("reference")  # type: ignore[arg-type]
            else:
                exclude_input = {"reference"}

        output = self.model_dump(
            mode=mode,
            by_alias=by_alias,
            exclude=exclude_input,
            exclude_none=exclude_none,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            context=context,
        )
        is_reference_user_excluded = isinstance(exclude, dict | set) and "reference" in exclude
        if self.reference is not None and not is_reference_user_excluded:
            # If the rules has a reference, we dump that separately with the as_reference flag set to True.
            # Unless the user has explicitly excluded the reference.
            output["Reference" if by_alias else "reference"] = self.reference.dump(
                mode=mode,
                by_alias=by_alias,
                exclude=exclude,
                exclude_none=exclude_none,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                entities_exclude_defaults=entities_exclude_defaults,
                as_reference=True,
            )
        return output


class SheetRow(SchemaModel):
    @abstractmethod
    def _identifier(self) -> tuple[Hashable, ...]:
        raise NotImplementedError()


T_SheetRow = TypeVar("T_SheetRow", bound=SheetRow)


class SheetList(list, MutableSequence[T_SheetRow]):
    @classmethod
    def __get_pydantic_core_schema__(cls, source: Any, handler: GetCoreSchemaHandler) -> core_schema.CoreSchema:
        if args := get_args(source):
            item_type = args[0]
        else:
            # Someone use SheetList without specifying the type
            raise TypeError("SheetList must be used with a type argument, e.g., SheetList[InformationProperty]")

        instance_schema = core_schema.is_instance_schema(cls)
        sequence_row_schema = handler.generate_schema(Sequence[item_type])  # type: ignore[valid-type]

        non_instance_schema = core_schema.no_info_after_validator_function(SheetList, sequence_row_schema)
        return core_schema.union_schema([instance_schema, non_instance_schema])

    def to_pandas(self, drop_na_columns: bool = True, include: list[str] | None = None) -> pd.DataFrame:
        """Converts ResourceDict to pandas DataFrame."""
        df = pd.DataFrame([entity.model_dump() for entity in self])
        if drop_na_columns:
            df = df.dropna(axis=1, how="all")
        if include is not None:
            df = df[include]
        return df

    def _repr_html_(self) -> str:
        """Returns HTML representation of ResourceDict."""
        return self.to_pandas(drop_na_columns=True)._repr_html_()  # type: ignore[operator]

    # Implemented to get correct type hints
    def __iter__(self) -> Iterator[T_SheetRow]:
        return super().__iter__()

    @overload
    def __getitem__(self, index: SupportsIndex) -> T_SheetRow: ...

    @overload
    def __getitem__(self, index: slice) -> SheetList[T_SheetRow]: ...

    def __getitem__(self, index: SupportsIndex | slice, /) -> T_SheetRow | SheetList[T_SheetRow]:
        if isinstance(index, slice):
            return SheetList[T_SheetRow](super().__getitem__(index))
        return super().__getitem__(index)


ExtensionCategoryType = Annotated[
    ExtensionCategory,
    PlainSerializer(
        lambda v: v.value if isinstance(v, ExtensionCategory) else v,
        return_type=str,
        when_used="unless-none",
    ),
    BeforeValidator(lambda v: ExtensionCategory(v) if isinstance(v, str) else v),
]


# Immutable such that this can be used as a key in a dictionary
class PropertyRef(BaseModel, frozen=True):
    class_: ClassEntityType = Field(alias="Class")
    property_: InformationPropertyType = Field(alias="Property")


class ClassRef(BaseModel, frozen=True):
    class_: ClassEntityType = Field(alias="Class")



from rdflib import Namespace

from cognite.neat._rules.importers._rdf._shared import make_metadata_compliant
from cognite.neat._rules.models import RoleTypes, SchemaCompleteness


def parse_imf_metadata(prefix: str = "pcaimf") -> dict:
    """Provide hardcoded IMF metadata to dict.

    Returns:
        Dictionary containing IMF metadata

    !!! note "Compliant IMF metadata"
        The current RDF provide IMF types as SHACL, but there are not any metadata describing
        the actual content.

    """

    raw_metadata = {
        "role": RoleTypes.information,
        "schema": SchemaCompleteness.partial,
        "prefix": prefix,
        "namespace": Namespace("https://posccaesar.org/imf/"),
        "version": None,
        "created": None,
        "updated": None,
        "title": "IMF_types",
        "description": "IMF - types",
        "creator": None,
        "rights": None,
        "license": None,
    }

    return make_metadata_compliant(raw_metadata)



from dataclasses import dataclass

from cognite.neat._issues import NeatError


@dataclass(unsafe_hash=True)
class WorkFlowMissingDataError(NeatError, ValueError):
    """In the workflow step {step_name} the following data is missing: {missing_data}."""

    step_name: str
    missing_data: frozenset[str]


@dataclass(unsafe_hash=True)
class WorkflowStepNotInitializedError(NeatError, RuntimeError):
    """Step {step_name} has not been initialized."""

    step_name: str


@dataclass(unsafe_hash=True)
class WorkflowConfigurationNotSetError(NeatError, RuntimeError):
    """The configuration variable '{config_variable}' is not set. Please set the configuration
    before running the workflow."""

    config_variable: str


@dataclass(unsafe_hash=True)
class WorkflowStepOutputError(NeatError, RuntimeError):
    """Object type {step_type} is not supported as step output.

    Step output must be of type DataContract or a FlowMessage.
    """

    step_type: str



import math
from collections.abc import Hashable
from dataclasses import dataclass, field
from typing import ClassVar

from pydantic import Field, field_serializer, field_validator

from cognite.neat._rules.models.data_types import DataType
from cognite.neat._rules.models.entities import ClassEntity, ClassEntityList

from ._base_input import InputComponent, InputRules
from ._base_rules import (
    BaseMetadata,
    BaseRules,
    RoleTypes,
    SheetList,
    SheetRow,
)
from ._types import ClassEntityType, InformationPropertyType, StrOrListType


class DomainMetadata(BaseMetadata):
    role: ClassVar[RoleTypes] = RoleTypes.domain_expert
    creator: StrOrListType

    def as_identifier(self) -> str:
        return "DomainRules"

    def get_prefix(self) -> str:
        return "domain"


class DomainProperty(SheetRow):
    class_: ClassEntityType = Field(alias="Class")
    property_: InformationPropertyType = Field(alias="Property")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    value_type: DataType | ClassEntity = Field(alias="Value Type")
    min_count: int | None = Field(alias="Min Count", default=None)
    max_count: int | float | None = Field(alias="Max Count", default=None)

    def _identifier(self) -> tuple[Hashable, ...]:
        return self.class_, self.property_

    @field_serializer("max_count", when_used="json-unless-none")
    def serialize_max_count(self, value: int | float | None) -> int | float | None | str:
        if isinstance(value, float) and math.isinf(value):
            return None
        return value

    @field_validator("max_count", mode="before")
    def parse_max_count(cls, value: int | float | None) -> int | float | None:
        if value is None:
            return float("inf")
        return value


class DomainClass(SheetRow):
    class_: ClassEntityType = Field(alias="Class")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(None, alias="Description")
    parent: ClassEntityList | None = Field(alias="Parent Class")

    def _identifier(self) -> tuple[Hashable, ...]:
        return (self.class_,)

    @field_serializer("parent", when_used="unless-none")
    def serialize_parent(self, value: list[ClassEntity]) -> str:
        return ",".join([str(entry) for entry in value])


class DomainRules(BaseRules):
    metadata: DomainMetadata = Field(alias="Metadata")
    properties: SheetList[DomainProperty] = Field(alias="Properties")
    classes: SheetList[DomainClass] | None = Field(None, alias="Classes")
    last: "DomainRules | None" = Field(None, alias="Last")
    reference: "DomainRules | None" = Field(None, alias="Reference")


@dataclass
class DomainInputMetadata(InputComponent[DomainMetadata]):
    creator: str

    @classmethod
    def _get_verified_cls(cls) -> type[DomainMetadata]:
        return DomainMetadata


@dataclass
class DomainInputProperty(InputComponent[DomainProperty]):
    class_: str
    property_: str
    value_type: str
    name: str | None = None
    description: str | None = None
    min_count: int | None = None
    max_count: int | float | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[DomainProperty]:
        return DomainProperty


@dataclass
class DomainInputClass(InputComponent[DomainClass]):
    class_: str
    name: str | None = None
    description: str | None = None
    parent: list[str] | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[DomainClass]:
        return DomainClass


@dataclass
class DomainInputRules(InputRules[DomainRules]):
    metadata: DomainInputMetadata
    properties: list[DomainInputProperty] = field(default_factory=list)
    classes: list[DomainInputClass] = field(default_factory=list)
    last: "DomainInputRules | None" = None
    reference: "DomainInputRules | None" = None

    @classmethod
    def _get_verified_cls(cls) -> type[DomainRules]:
        return DomainRules



from abc import ABC, abstractmethod
from collections.abc import MutableSequence
from typing import Generic, TypeVar

from cognite.neat._issues import IssueList, NeatError
from cognite.neat._issues.errors import NeatTypeError, NeatValueError
from cognite.neat._rules._shared import (
    InputRules,
    JustRules,
    MaybeRules,
    OutRules,
    Rules,
    VerifiedRules,
)

T_RulesIn = TypeVar("T_RulesIn", bound=Rules)
T_RulesOut = TypeVar("T_RulesOut", bound=Rules)


class RulesTransformer(ABC, Generic[T_RulesIn, T_RulesOut]):
    """This is the base class for all rule transformers.

    Note transformers follow the functional pattern Monad
    https://en.wikipedia.org/wiki/Monad_(functional_programming)
    """

    @abstractmethod
    def transform(self, rules: T_RulesIn | OutRules[T_RulesIn]) -> OutRules[T_RulesOut]:
        """Transform the input rules into the output rules."""
        raise NotImplementedError()

    def try_transform(self, rules: MaybeRules[T_RulesIn]) -> MaybeRules[T_RulesOut]:
        """Try to transform the input rules into the output rules."""
        try:
            result = self.transform(rules)
        except NeatError:
            # Any error caught during transformation will be returned as issues
            return MaybeRules(None, rules.issues)
        issues = IssueList(rules.issues, title=rules.issues.title)
        if isinstance(result, MaybeRules):
            issues.extend(result.issues)
        return MaybeRules(result.get_rules(), issues)

    @classmethod
    def _to_rules(cls, rules: T_RulesIn | OutRules[T_RulesIn]) -> T_RulesIn:
        if isinstance(rules, JustRules):
            return rules.rules
        elif isinstance(rules, MaybeRules):
            if rules.rules is None:
                raise NeatValueError("Rules is missing cannot convert")
            return rules.rules
        elif isinstance(rules, VerifiedRules | InputRules):
            return rules  # type: ignore[return-value]
        else:
            raise NeatTypeError(f"Unsupported type: {type(rules)}")


class RulesPipeline(list, MutableSequence[RulesTransformer], Generic[T_RulesIn, T_RulesOut]):
    def transform(self, rules: T_RulesIn | OutRules[T_RulesIn]) -> OutRules[T_RulesOut]:
        """Transform the input rules into the output rules."""
        for transformer in self:
            rules = transformer.transform(rules)
        return rules  # type: ignore[return-value]

    def try_transform(self, rules: MaybeRules[T_RulesIn]) -> MaybeRules[T_RulesOut]:
        """Try to transform the input rules into the output rules."""
        for transformer in self:
            rules = transformer.try_transform(rules)
        return rules  # type: ignore[return-value]

    def run(self, rules: T_RulesIn | OutRules[T_RulesIn]) -> T_RulesOut:
        """Run the pipeline from the input rules to the output rules."""
        output = self.transform(rules)
        if isinstance(output, MaybeRules):
            if output.rules is None:
                raise NeatValueError(f"Rule transformation failed: {output.issues}")
            return output.rules
        elif isinstance(output, JustRules):
            return output.rules
        else:
            raise NeatTypeError(f"Rule transformation failed: {output}")



from collections import defaultdict
from collections.abc import Callable, Iterable, Set
from datetime import datetime, timezone
from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes import Relationship, RelationshipList
from rdflib import RDF, Literal, Namespace

from cognite.neat._graph.models import Triple
from cognite.neat._utils.auxiliary import create_sha256_hash

from ._base import DEFAULT_SKIP_METADATA_VALUES, ClassicCDFBaseExtractor, InstanceIdPrefix
from ._labels import LabelsExtractor


class RelationshipsExtractor(ClassicCDFBaseExtractor[Relationship]):
    """Extract data from Cognite Data Fusions Relationships into Neat.

    Args:
        items (Iterable[Relationship]): An iterable of items.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
        to_type (Callable[[Relationship], str | None], optional): A function to convert an item to a type.
            Defaults to None. If None or if the function returns None, the asset will be set to the default type.
        total (int, optional): The total number of items to load. If passed, you will get a progress bar if rich
            is installed. Defaults to None.
        limit (int, optional): The maximal number of items to load. Defaults to None. This is typically used for
            testing setup of the extractor. For example, if you are extracting 100 000 assets, you might want to
            limit the extraction to 1000 assets to test the setup.
        unpack_metadata (bool, optional): Whether to unpack metadata. Defaults to False, which yields the metadata as
            a JSON string.
        skip_metadata_values (set[str] | frozenset[str] | None, optional): If you are unpacking metadata, then
           values in this set will be skipped.
    """

    _default_rdf_type = "Relationship"

    def __init__(
        self,
        items: Iterable[Relationship],
        namespace: Namespace | None = None,
        to_type: Callable[[Relationship], str | None] | None = None,
        total: int | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
    ):
        super().__init__(
            items,
            namespace=namespace,
            to_type=to_type,
            total=total,
            limit=limit,
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
        )
        # This is used by the ClassicExtractor to log the target nodes, such
        # that it can extract them.
        # It is private to avoid exposing it to the user.
        self._log_target_nodes = False
        self._target_external_ids_by_type: dict[InstanceIdPrefix, set[str]] = defaultdict(set)

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
        to_type: Callable[[Relationship], str | None] | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
    ):
        return cls(
            client.relationships(data_set_external_ids=data_set_external_id),
            namespace=namespace,
            to_type=to_type,
            limit=limit,
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
        )

    @classmethod
    def from_file(
        cls,
        file_path: str,
        namespace: Namespace | None = None,
        to_type: Callable[[Relationship], str | None] | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
    ):
        relationships = RelationshipList.load(Path(file_path).read_text())
        return cls(
            relationships,
            namespace=namespace,
            total=len(relationships),
            to_type=to_type,
            limit=limit,
            unpack_metadata=unpack_metadata,
            skip_metadata_values=skip_metadata_values,
        )

    def _item2triples(self, relationship: Relationship) -> list[Triple]:
        """Converts an asset to triples."""

        if relationship.external_id and relationship.source_external_id and relationship.target_external_id:
            if self._log_target_nodes and relationship.target_type and relationship.target_external_id:
                self._target_external_ids_by_type[InstanceIdPrefix.from_str(relationship.target_type)].add(
                    relationship.target_external_id
                )

            # relationships do not have an internal id, so we generate one
            id_ = self.namespace[f"{InstanceIdPrefix.relationship}{create_sha256_hash(relationship.external_id)}"]

            type_ = self._get_rdf_type(relationship)
            # Set rdf type
            triples: list[Triple] = [(id_, RDF.type, self.namespace[type_])]

            # Set source and target types
            if source_type := relationship.source_type:
                triples.append(
                    (
                        id_,
                        self.namespace.source_type,
                        self.namespace[source_type.title()],
                    )
                )

            if target_type := relationship.target_type:
                triples.append(
                    (
                        id_,
                        self.namespace.target_type,
                        self.namespace[target_type.title()],
                    )
                )

            # Create attributes

            triples.append((id_, self.namespace.external_id, Literal(relationship.external_id)))

            triples.append(
                (
                    id_,
                    self.namespace.source_external_id,
                    Literal(relationship.source_external_id),
                )
            )

            triples.append(
                (
                    id_,
                    self.namespace.target_external_id,
                    Literal(relationship.target_external_id),
                )
            )

            if relationship.start_time:
                triples.append(
                    (
                        id_,
                        self.namespace.start_time,
                        Literal(datetime.fromtimestamp(relationship.start_time / 1000, timezone.utc)),
                    )
                )

            if relationship.end_time:
                triples.append(
                    (
                        id_,
                        self.namespace.end_time,
                        Literal(datetime.fromtimestamp(relationship.end_time / 1000, timezone.utc)),
                    )
                )

            if relationship.created_time:
                triples.append(
                    (
                        id_,
                        self.namespace.created_time,
                        Literal(datetime.fromtimestamp(relationship.created_time / 1000, timezone.utc)),
                    )
                )

            if relationship.last_updated_time:
                triples.append(
                    (
                        id_,
                        self.namespace.last_updated_time,
                        Literal(datetime.fromtimestamp(relationship.last_updated_time / 1000, timezone.utc)),
                    )
                )

            if relationship.confidence:
                triples.append(
                    (
                        id_,
                        self.namespace.confidence,
                        Literal(relationship.confidence),
                    )
                )

            if relationship.labels:
                for label in relationship.labels:
                    # external_id can create ill-formed URIs, so we create websafe URIs
                    # since labels do not have internal ids, we use the external_id as the id
                    triples.append(
                        (
                            id_,
                            self.namespace.label,
                            self.namespace[f"{InstanceIdPrefix.label}{LabelsExtractor._label_id(label)}"],
                        )
                    )

            # Create connection
            if relationship.data_set_id:
                triples.append(
                    (
                        id_,
                        self.namespace.dataset,
                        self.namespace[f"{InstanceIdPrefix.data_set}{relationship.data_set_id}"],
                    )
                )

            return triples
        return []



from rdflib import Graph, Namespace

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._rules.importers._rdf._shared import make_metadata_compliant
from cognite.neat._rules.models import RoleTypes, SchemaCompleteness
from cognite.neat._utils.collection_ import remove_none_elements_from_set
from cognite.neat._utils.rdf_ import convert_rdflib_content


def parse_owl_metadata(graph: Graph) -> dict:
    """Parse owl metadata from graph to dict.

    Args:
        graph: Graph containing owl metadata

    Returns:
        Dictionary containing owl metadata

    !!! note "Compliant OWL metadata"
        This makes the method very opinionated, but results in a compliant metadata.


    """
    # TODO: Move dataframe to dict representation

    query = f"""SELECT ?namespace ?prefix ?version ?created ?updated ?title ?description ?creator ?rights ?license
    WHERE {{
        ?namespace a owl:Ontology .
        OPTIONAL {{?namespace owl:versionInfo ?version }}.
        OPTIONAL {{?namespace dcterms:creator ?creator }}.
        OPTIONAL {{?namespace <{DEFAULT_NAMESPACE.prefix}> ?prefix }}.
        OPTIONAL {{?namespace dcterms:title|rdfs:label|skos:prefLabel ?title }}.
        OPTIONAL {{?namespace dcterms:modified ?updated }}.
        OPTIONAL {{?namespace dcterms:created ?created }}.
        OPTIONAL {{?namespace dcterms:description ?description }}.
        OPTIONAL {{?namespace dcterms:rights|dc:rights ?rights }}.

        OPTIONAL {{?namespace dcterms:license|dc:license ?license }}.
        FILTER (!isBlank(?namespace))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "en"))
        FILTER (!bound(?title) || LANG(?title) = "" || LANGMATCHES(LANG(?title), "en"))
    }}
    """

    results = [{item for item in sublist} for sublist in list(zip(*graph.query(query), strict=True))]

    raw_metadata = convert_rdflib_content(
        {
            "role": RoleTypes.information,
            "schema": SchemaCompleteness.partial,
            "prefix": results[1].pop(),
            "namespace": Namespace(results[0].pop()),
            "version": results[2].pop(),
            "created": results[3].pop(),
            "updated": results[4].pop(),
            "title": results[5].pop(),
            "description": results[6].pop(),
            "creator": (
                ", ".join(remove_none_elements_from_set(results[7]))
                if remove_none_elements_from_set(results[7])
                else None
            ),
            "rights": results[8].pop(),
            "license": results[9].pop(),
        }
    )

    return make_metadata_compliant(raw_metadata)



from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes import (
    Asset,
    AssetUpdate,
    Relationship,
    RelationshipUpdate,
)
from cognite.client.data_classes.data_modeling import EdgeApply, NodeApply

from cognite.neat._rules.models import (
    AssetRules,
    DMSRules,
    DomainRules,
    InformationRules,
)
from cognite.neat._store import NeatGraphStore
from cognite.neat._workflows.steps.step_model import DataContract


class MultiRuleData(DataContract):
    domain: DomainRules | None = None
    information: InformationRules | None = None
    asset: AssetRules | None = None
    dms: DMSRules | None = None

    @classmethod
    def from_rules(cls, rules: DomainRules | InformationRules | AssetRules | DMSRules):
        if isinstance(rules, DomainRules):
            return cls(domain=rules)
        elif isinstance(rules, InformationRules):
            return cls(information=rules)
        elif isinstance(rules, AssetRules):
            return cls(asset=rules)
        elif isinstance(rules, DMSRules):
            return cls(dms=rules)
        else:
            raise ValueError(f"Unsupported rules type {type(rules)}")


class PathData(DataContract):
    """
    This represents the path to the excel file.

    Args:
        excel_file_path: The path to the excel file.
    """

    excel_file_path: Path


class NeatGraph(DataContract):
    """
    This represents the neat graph.

    Args:
        graph: The neat graph store.
    """

    graph: NeatGraphStore


class ClientData(DataContract):
    """
    This represents an instantiated CogniteClient object.

    Args:
        client: The CogniteClient object.
    """

    client: CogniteClient


class CategorizedAssets(DataContract):
    """ "
    This represents the categorized assets.

    Args:
        assets: The categorized assets.
    """

    assets: dict[str, list[Asset | AssetUpdate]]


class CategorizedRelationships(DataContract):
    """
    This represents the categorized relationships.

    Args:
        relationships: The categorized relationships.

    """

    relationships: dict[str, list[Relationship] | list[RelationshipUpdate]]


class Nodes(DataContract):
    """
    This represents nodes.

    Args:
        nodes: list of nodes.
    """

    nodes: list[NodeApply]


class Edges(DataContract):
    """
    This represents edges.

    Args:
        edges: list of edges.
    """

    edges: list[EdgeApply]



from typing import Literal, cast

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.neat import _version
from cognite.neat._issues import IssueList
from cognite.neat._rules import importers
from cognite.neat._rules._shared import ReadRules
from cognite.neat._rules.importers._rdf._base import DEFAULT_NON_EXISTING_NODE_TYPE
from cognite.neat._rules.models import DMSRules
from cognite.neat._rules.models.data_types import AnyURI
from cognite.neat._rules.models.entities._single_value import UnknownEntity
from cognite.neat._rules.models.information._rules import InformationRules
from cognite.neat._rules.models.information._rules_input import InformationInputRules
from cognite.neat._rules.transformers import ConvertToRules, VerifyAnyRules

from ._inspect import InspectAPI
from ._prepare import PrepareAPI
from ._read import ReadAPI
from ._set import SetAPI
from ._show import ShowAPI
from ._state import SessionState
from ._to import ToAPI
from .exceptions import intercept_session_exceptions


@intercept_session_exceptions
class NeatSession:
    def __init__(
        self,
        client: CogniteClient | None = None,
        storage: Literal["memory", "oxigraph"] = "memory",
        verbose: bool = True,
    ) -> None:
        self._client = client
        self._verbose = verbose
        self._state = SessionState(store_type=storage)
        self.read = ReadAPI(self._state, client, verbose)
        self.to = ToAPI(self._state, client, verbose)
        self.prepare = PrepareAPI(self._state, verbose)
        self.show = ShowAPI(self._state)
        self.set = SetAPI(self._state, verbose)
        self.inspect = InspectAPI(self._state)

    @property
    def version(self) -> str:
        return _version.__version__

    def verify(self) -> IssueList:
        output = VerifyAnyRules("continue").try_transform(self._state.input_rule)
        if output.rules:
            self._state.verified_rules.append(output.rules)
            if isinstance(output.rules, InformationRules):
                self._state.store.add_rules(output.rules)
        output.issues.action = "verify"
        self._state.issue_lists.append(output.issues)
        if output.issues:
            print("You can inspect the issues with the .inspect.issues(...) method.")
        return output.issues

    def convert(self, target: Literal["dms"]) -> None:
        converted = ConvertToRules(DMSRules).transform(self._state.last_verified_rule)
        self._state.verified_rules.append(converted.rules)
        if self._verbose:
            print(f"Rules converted to {target}")

    def infer(
        self,
        model_id: dm.DataModelId | tuple[str, str, str] = (
            "neat_space",
            "NeatInferredDataModel",
            "v1",
        ),
        non_existing_node_type: UnknownEntity | AnyURI = DEFAULT_NON_EXISTING_NODE_TYPE,
    ) -> IssueList:
        """Data model inference from instances.

        Args:
            model_id: The ID of the inferred data model.
            non_existing_node_type: The type of node to use when type of node is not possible to determine.
        """

        model_id = dm.DataModelId.load(model_id)

        input_rules: ReadRules = importers.InferenceImporter.from_graph_store(
            store=self._state.store,
            non_existing_node_type=non_existing_node_type,
        ).to_rules()

        if model_id.space:
            cast(InformationInputRules, input_rules.rules).metadata.prefix = model_id.space
        if model_id.external_id:
            cast(InformationInputRules, input_rules.rules).metadata.name = model_id.external_id

        if model_id.version:
            cast(InformationInputRules, input_rules.rules).metadata.version = model_id.version

        self.read.rdf._store_rules(self._state.store, input_rules, "Data Model Inference")
        return input_rules.issues

    def _repr_html_(self) -> str:
        state = self._state
        if not state.has_store and not state.input_rules:
            return "<strong>Empty session</strong>. Get started by reading something with the <em>.read</em> attribute."

        output = []
        if state.input_rules and not state.verified_rules:
            output.append(f"<H2>Unverified Data Model</H2><br />{state.input_rule.rules._repr_html_()}")  # type: ignore

        if state.verified_rules:
            output.append(f"<H2>Verified Data Model</H2><br />{state.last_verified_rule._repr_html_()}")  # type: ignore

        if state.has_store:
            output.append(f"<H2>Instances</H2> {state.store._repr_html_()}")

        return "<br />".join(output)



from cognite.neat._constants import DMS_LISTABLE_PROPERTY_LIMIT
from cognite.neat._rules.models.dms import DMSProperty, DMSRules, DMSView
from cognite.neat._rules.models.entities import ReferenceEntity, ViewEntity

from ._base import BaseAnalysis


class DMSAnalysis(BaseAnalysis[DMSRules, DMSView, DMSProperty, ViewEntity, str]):
    """Assumes analysis over only the complete schema"""

    def _get_classes(self) -> list[DMSView]:
        return list(self.rules.views)

    def _get_properties(self) -> list[DMSProperty]:
        return list(self.rules.properties)

    def _get_reference(self, class_or_property: DMSView | DMSProperty) -> ReferenceEntity | None:
        return class_or_property.reference if isinstance(class_or_property.reference, ReferenceEntity) else None

    def _get_cls_entity(self, class_: DMSView | DMSProperty) -> ViewEntity:
        return class_.view

    def _get_prop_entity(self, property_: DMSProperty) -> str:
        return property_.property_

    def _get_cls_parents(self, class_: DMSView) -> list[ViewEntity] | None:
        return list(class_.implements) if class_.implements else None

    def _get_reference_rules(self) -> DMSRules | None:
        return self.rules.reference

    @classmethod
    def _set_cls_entity(cls, property_: DMSProperty, class_: ViewEntity) -> None:
        property_.view = class_

    def _get_object(self, property_: DMSProperty) -> ViewEntity | None:
        return property_.value_type if isinstance(property_.value_type, ViewEntity) else None

    def _get_max_occurrence(self, property_: DMSProperty) -> int | float | None:
        return DMS_LISTABLE_PROPERTY_LIMIT if property_.is_list else 1

    def subset_rules(self, desired_classes: set[ViewEntity]) -> DMSRules:
        raise NotImplementedError()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, TypeAlias, TypeVar

from cognite.neat._issues import IssueList
from cognite.neat._rules.models import (
    AssetRules,
    DMSRules,
    DomainRules,
    InformationRules,
)
from cognite.neat._rules.models.asset._rules_input import AssetInputRules
from cognite.neat._rules.models.dms._rules_input import DMSInputRules
from cognite.neat._rules.models.information._rules_input import InformationInputRules

VerifiedRules: TypeAlias = DomainRules | InformationRules | DMSRules | AssetRules
InputRules: TypeAlias = AssetInputRules | DMSInputRules | InformationInputRules
Rules: TypeAlias = (
    AssetInputRules | DMSInputRules | InformationInputRules | DomainRules | InformationRules | DMSRules | AssetRules
)
T_Rules = TypeVar("T_Rules", bound=Rules)
T_VerifiedRules = TypeVar("T_VerifiedRules", bound=VerifiedRules)
T_InputRules = TypeVar("T_InputRules", bound=InputRules)


@dataclass
class OutRules(Generic[T_Rules], ABC):
    """This is a base class for all rule states."""

    @abstractmethod
    def get_rules(self) -> T_Rules | None:
        """Get the rules from the state."""
        raise NotImplementedError()


@dataclass
class JustRules(OutRules[T_Rules]):
    """This represents a rule that exists"""

    rules: T_Rules

    def get_rules(self) -> T_Rules:
        return self.rules


@dataclass
class MaybeRules(OutRules[T_Rules]):
    """This represents a rule that may or may not exist"""

    rules: T_Rules | None
    issues: IssueList

    def get_rules(self) -> T_Rules | None:
        return self.rules


@dataclass
class ReadRules(MaybeRules[T_Rules]):
    """This represents a rule that does not exist"""

    read_context: dict[str, Any]



from pathlib import Path
from typing import Any, Literal, overload

from cognite.client import CogniteClient

from cognite.neat._graph import loaders
from cognite.neat._issues import IssueList, catch_warnings
from cognite.neat._rules import exporters
from cognite.neat._session._wizard import space_wizard
from cognite.neat._utils.upload import UploadResultCore

from ._state import SessionState
from .exceptions import intercept_session_exceptions


@intercept_session_exceptions
class ToAPI:
    def __init__(self, state: SessionState, client: CogniteClient | None, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose
        self.cdf = CDFToAPI(state, client, verbose)

    def excel(
        self,
        io: Any,
    ) -> None:
        exporter = exporters.ExcelExporter()
        exporter.export_to_file(self._state.last_verified_rule, Path(io))
        return None

    @overload
    def yaml(self, io: None) -> str: ...

    @overload
    def yaml(self, io: Any) -> None: ...

    def yaml(self, io: Any | None = None) -> str | None:
        exporter = exporters.YAMLExporter()
        if io is None:
            return exporter.export(self._state.last_verified_rule)

        exporter.export_to_file(self._state.last_verified_rule, Path(io))
        return None


@intercept_session_exceptions
class CDFToAPI:
    def __init__(self, state: SessionState, client: CogniteClient | None, verbose: bool) -> None:
        self._client = client
        self._state = state
        self._verbose = verbose

    def instances(self, space: str | None = None):
        if not self._state.last_verified_dms_rules:
            raise ValueError("No verified DMS data model available")

        loader = loaders.DMSLoader.from_rules(
            self._state.last_verified_dms_rules, self._state.store, space_wizard(space=space)
        )

        if not self._client:
            raise ValueError("No client provided!")

        return loader.load_into_cdf(self._client)

    def data_model(
        self,
        existing_handling: Literal["fail", "skip", "update", "force"] = "skip",
        dry_run: bool = False,
        fallback_one_by_one: bool = False,
    ):
        """Export the verified DMS data model to CDF.

        Args:
            existing_handling: How to handle if component of data model exists. Defaults to "skip".
            dry_run: If True, no changes will be made to CDF. Defaults to False.
            fallback_one_by_one: If True, will fall back to one-by-one upload if batch upload fails. Defaults to False.

        ... note::

        - "fail": If any component already exists, the export will fail.
        - "skip": If any component already exists, it will be skipped.
        - "update": If any component already exists, it will be updated.
        - "force": If any component already exists, it will be deleted and recreated.

        """
        if not self._state.last_verified_dms_rules:
            raise ValueError("No verified DMS data model available")

        exporter = exporters.DMSExporter(existing_handling=existing_handling)

        if not self._client:
            raise ValueError("No client provided!")

        conversion_issues = IssueList(action="to.cdf.data_model")
        with catch_warnings(conversion_issues):
            result = exporter.export_to_cdf(
                self._state.last_verified_dms_rules, self._client, dry_run, fallback_one_by_one
            )
        result.insert(0, UploadResultCore(name="schema", issues=conversion_issues))
        self._state.outcome.append(result)
        print("You can inspect the details with the .inspect.outcome(...) method.")
        return result



from ._classic_cdf import (
    AddAssetDepth,
    AssetEventConnector,
    AssetFileConnector,
    AssetRelationshipConnector,
    AssetSequenceConnector,
    AssetTimeSeriesConnector,
    RelationshipToSchemaTransformer,
)
from ._rdfpath import AddSelfReferenceProperty
from ._value_type import SplitMultiValueProperty

__all__ = [
    "AddAssetDepth",
    "AssetTimeSeriesConnector",
    "AssetSequenceConnector",
    "AssetFileConnector",
    "AssetEventConnector",
    "AssetRelationshipConnector",
    "AddSelfReferenceProperty",
    "SplitMultiValueProperty",
    "RelationshipToSchemaTransformer",
]

Transformers = (
    AddAssetDepth
    | AssetTimeSeriesConnector
    | AssetSequenceConnector
    | AssetFileConnector
    | AssetEventConnector
    | AssetRelationshipConnector
    | AddSelfReferenceProperty
    | SplitMultiValueProperty
    | RelationshipToSchemaTransformer
)



import math
import sys
import warnings
from collections.abc import Hashable
from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar, Literal

import pandas as pd
from cognite.client import data_modeling as dm
from pydantic import Field, field_serializer, field_validator, model_validator
from pydantic_core.core_schema import SerializationInfo, ValidationInfo

from cognite.neat._issues import MultiValueError
from cognite.neat._issues.warnings import (
    PrincipleMatchingSpaceAndVersionWarning,
    PrincipleSolutionBuildsOnEnterpriseWarning,
)
from cognite.neat._rules.models._base_rules import (
    BaseMetadata,
    BaseRules,
    DataModelType,
    ExtensionCategory,
    RoleTypes,
    SchemaCompleteness,
    SheetList,
    SheetRow,
)
from cognite.neat._rules.models._types import (
    ClassEntityType,
    ContainerEntityType,
    DataModelExternalIdType,
    DmsPropertyType,
    InformationPropertyType,
    SpaceType,
    StrListType,
    VersionType,
    ViewEntityType,
)
from cognite.neat._rules.models.data_types import DataType
from cognite.neat._rules.models.entities import (
    ContainerEntityList,
    DMSEntity,
    DMSNodeEntity,
    DMSUnknownEntity,
    EdgeEntity,
    Entity,
    HasDataFilter,
    NodeTypeFilter,
    RawFilter,
    ReferenceEntity,
    ReverseConnectionEntity,
    URLEntity,
    ViewEntity,
    ViewEntityList,
)

from ._schema import DMSSchema

if TYPE_CHECKING:
    pass

if sys.version_info >= (3, 11):
    pass
else:
    pass

_DEFAULT_VERSION = "1"


class DMSMetadata(BaseMetadata):
    role: ClassVar[RoleTypes] = RoleTypes.dms
    data_model_type: DataModelType = Field(DataModelType.enterprise, alias="dataModelType")
    schema_: SchemaCompleteness = Field(alias="schema")
    extension: ExtensionCategory = ExtensionCategory.addition
    space: SpaceType
    name: str | None = Field(
        None,
        description="Human readable name of the data model",
        min_length=1,
        max_length=255,
    )
    description: str | None = Field(None, min_length=1, max_length=1024)
    external_id: DataModelExternalIdType = Field(alias="externalId")
    version: VersionType
    creator: StrListType
    created: datetime = Field(
        description=("Date of the data model creation"),
    )
    updated: datetime = Field(
        description=("Date of the data model update"),
    )

    @field_validator("*", mode="before")
    def strip_string(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.strip()
        return value

    @field_serializer("schema_", "extension", "data_model_type", when_used="always")
    def as_string(self, value: SchemaCompleteness | ExtensionCategory | DataModelType) -> str:
        return str(value)

    @field_validator("schema_", mode="plain")
    def as_enum_schema(cls, value: str) -> SchemaCompleteness:
        return SchemaCompleteness(value.strip())

    @field_validator("extension", mode="plain")
    def as_enum_extension(cls, value: str) -> ExtensionCategory:
        return ExtensionCategory(value.strip())

    @field_validator("data_model_type", mode="plain")
    def as_enum_model_type(cls, value: str) -> DataModelType:
        return DataModelType(value.strip())

    @field_validator("description", mode="before")
    def nan_as_none(cls, value):
        if isinstance(value, float) and math.isnan(value):
            return None
        return value

    def as_space(self) -> dm.SpaceApply:
        return dm.SpaceApply(
            space=self.space,
        )

    def as_data_model_id(self) -> dm.DataModelId:
        return dm.DataModelId(space=self.space, external_id=self.external_id, version=self.version)

    def as_data_model(self) -> dm.DataModelApply:
        suffix = f"Creator: {', '.join(self.creator)}"
        if self.description:
            description = f"{self.description} Creator: {', '.join(self.creator)}"
        else:
            description = suffix

        return dm.DataModelApply(
            space=self.space,
            external_id=self.external_id,
            name=self.name or None,
            version=self.version or "missing",
            description=description,
            views=[],
        )

    def as_identifier(self) -> str:
        return repr(self.as_data_model_id())

    def get_prefix(self) -> str:
        return self.space


def _metadata(context: Any) -> DMSMetadata | None:
    if isinstance(context, dict) and isinstance(context.get("metadata"), DMSMetadata):
        return context["metadata"]
    return None


class DMSProperty(SheetRow):
    view: ViewEntityType = Field(alias="View")
    view_property: DmsPropertyType = Field(alias="View Property")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    connection: Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None = Field(None, alias="Connection")
    value_type: DataType | ViewEntity | DMSUnknownEntity = Field(alias="Value Type")
    nullable: bool | None = Field(default=None, alias="Nullable")
    immutable: bool | None = Field(default=None, alias="Immutable")
    is_list: bool | None = Field(default=None, alias="Is List")
    default: str | int | dict | None = Field(None, alias="Default")
    reference: URLEntity | ReferenceEntity | None = Field(default=None, alias="Reference", union_mode="left_to_right")
    container: ContainerEntityType | None = Field(None, alias="Container")
    container_property: DmsPropertyType | None = Field(None, alias="Container Property")
    index: StrListType | None = Field(None, alias="Index")
    constraint: StrListType | None = Field(None, alias="Constraint")
    class_: ClassEntityType = Field(alias="Class (linage)")
    property_: InformationPropertyType = Field(alias="Property (linage)")

    def _identifier(self) -> tuple[Hashable, ...]:
        return self.view, self.view_property

    @field_validator("nullable")
    def direct_relation_must_be_nullable(cls, value: Any, info: ValidationInfo) -> None:
        if info.data.get("connection") == "direct" and value is False:
            raise ValueError("Direct relation must be nullable")
        return value

    @field_validator("value_type", mode="after")
    def connections_value_type(
        cls, value: EdgeEntity | ViewEntity | DMSUnknownEntity, info: ValidationInfo
    ) -> DataType | EdgeEntity | ViewEntity | DMSUnknownEntity:
        if (connection := info.data.get("connection")) is None:
            return value
        if connection == "direct" and not isinstance(value, ViewEntity | DMSUnknownEntity):
            raise ValueError(f"Direct relation must have a value type that points to a view, got {value}")
        elif isinstance(connection, EdgeEntity) and not isinstance(value, ViewEntity):
            raise ValueError(f"Edge connection must have a value type that points to a view, got {value}")
        elif isinstance(connection, ReverseConnectionEntity) and not isinstance(value, ViewEntity):
            raise ValueError(f"Reverse connection must have a value type that points to a view, got {value}")
        return value

    @field_serializer("reference", when_used="always")
    def set_reference(self, value: Any, info: SerializationInfo) -> str | None:
        if isinstance(info.context, dict) and info.context.get("as_reference") is True:
            return str(
                ReferenceEntity(
                    prefix=self.view.prefix,
                    suffix=self.view.suffix,
                    version=self.view.version,
                    property=self.view_property,
                )
            )
        return str(value) if value is not None else None

    @field_serializer("value_type", when_used="always")
    def as_dms_type(self, value_type: DataType | EdgeEntity | ViewEntity, info: SerializationInfo) -> str:
        if isinstance(value_type, DataType):
            return value_type._suffix_extra_args(value_type.dms._type)
        elif isinstance(value_type, EdgeEntity | ViewEntity) and (metadata := _metadata(info.context)):
            return value_type.dump(space=metadata.space, version=metadata.version)
        return str(value_type)

    @field_serializer("view", "container", "class_", when_used="unless-none")
    def remove_default_space(self, value: str, info: SerializationInfo) -> str:
        if (metadata := _metadata(info.context)) and isinstance(value, Entity):
            if info.field_name == "container" and info.context.get("as_reference") is True:
                # When dumping as reference, the container should keep the default space for easy copying
                # over to user sheets.
                return value.dump()
            return value.dump(prefix=metadata.space, version=metadata.version)
        return str(value)

    @field_serializer("connection", when_used="unless-none")
    def remove_defaults(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, Entity) and (metadata := _metadata(info.context)):
            default_type = f"{self.view.external_id}.{self.view_property}"
            if isinstance(value, EdgeEntity) and value.edge_type and value.edge_type.space != metadata.space:
                default_type = f"{metadata.space}{default_type}"
            return value.dump(space=metadata.space, version=metadata.version, type=default_type)
        return str(value)


class DMSContainer(SheetRow):
    container: ContainerEntityType = Field(alias="Container")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    reference: URLEntity | ReferenceEntity | None = Field(alias="Reference", default=None, union_mode="left_to_right")
    constraint: ContainerEntityList | None = Field(None, alias="Constraint")
    used_for: Literal["node", "edge", "all"] | None = Field("all", alias="Used For")
    class_: ClassEntityType = Field(alias="Class (linage)")

    def _identifier(self) -> tuple[Hashable, ...]:
        return (self.container,)

    def as_container(self) -> dm.ContainerApply:
        container_id = self.container.as_id()
        constraints: dict[str, dm.Constraint] = {}
        for constraint in self.constraint or []:
            requires = dm.RequiresConstraint(constraint.as_id())
            constraints[f"{constraint.space}_{constraint.external_id}"] = requires

        return dm.ContainerApply(
            space=container_id.space,
            external_id=container_id.external_id,
            name=self.name or None,
            description=self.description,
            constraints=constraints or None,
            properties={},
            used_for=self.used_for,
        )

    @field_serializer("reference", when_used="always")
    def set_reference(self, value: Any, info: SerializationInfo) -> str | None:
        if isinstance(info.context, dict) and info.context.get("as_reference") is True:
            return self.container.dump()
        return str(value) if value is not None else None

    @field_serializer("container", "class_", when_used="unless-none")
    def remove_default_space(self, value: Any, info: SerializationInfo) -> str:
        if metadata := _metadata(info.context):
            if isinstance(value, DMSEntity):
                return value.dump(space=metadata.space, version=metadata.version)
            elif isinstance(value, Entity):
                return value.dump(prefix=metadata.space, version=metadata.version)
        return str(value)

    @field_serializer("constraint", when_used="unless-none")
    def remove_default_spaces(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, list) and (metadata := _metadata(info.context)):
            return ",".join(
                constraint.dump(space=metadata.space, version=metadata.version)
                if isinstance(constraint, DMSEntity)
                else str(constraint)
                for constraint in value
            )
        return ",".join(str(value) for value in value)


class DMSView(SheetRow):
    view: ViewEntityType = Field(alias="View")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)
    implements: ViewEntityList | None = Field(None, alias="Implements")
    reference: URLEntity | ReferenceEntity | None = Field(alias="Reference", default=None, union_mode="left_to_right")
    filter_: HasDataFilter | NodeTypeFilter | RawFilter | None = Field(None, alias="Filter")
    in_model: bool = Field(True, alias="In Model")
    class_: ClassEntityType = Field(alias="Class (linage)")

    def _identifier(self) -> tuple[Hashable, ...]:
        return (self.view,)

    @field_serializer("reference", when_used="always")
    def set_reference(self, value: Any, info: SerializationInfo) -> str | None:
        if isinstance(info.context, dict) and info.context.get("as_reference") is True:
            return self.view.dump()
        return str(value) if value is not None else None

    @field_serializer("view", "class_", when_used="unless-none")
    def remove_default_space(self, value: Any, info: SerializationInfo) -> str:
        if (metadata := _metadata(info.context)) and isinstance(value, Entity):
            return value.dump(prefix=metadata.space, version=metadata.version)
        return str(value)

    @field_serializer("implements", when_used="unless-none")
    def remove_default_spaces(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, list) and (metadata := _metadata(info.context)):
            return ",".join(
                parent.dump(space=metadata.space, version=metadata.version)
                if isinstance(parent, DMSEntity)
                else str(parent)
                for parent in value
            )
        return ",".join(str(value) for value in value) if isinstance(value, list) else value

    def as_view(self) -> dm.ViewApply:
        view_id = self.view.as_id()
        implements = [parent.as_id() for parent in self.implements or []] or None
        if implements is None and isinstance(self.reference, ReferenceEntity):
            # Fallback to the reference if no implements are provided
            parent = self.reference.as_view_id()
            if (parent.space, parent.external_id) != (view_id.space, view_id.external_id):
                implements = [parent]

        return dm.ViewApply(
            space=view_id.space,
            external_id=view_id.external_id,
            version=view_id.version or _DEFAULT_VERSION,
            name=self.name or None,
            description=self.description,
            implements=implements,
            properties={},
        )


class DMSNode(SheetRow):
    node: DMSNodeEntity = Field(alias="Node")
    usage: Literal["type", "collection"] = Field(alias="Usage")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)

    def _identifier(self) -> tuple[Hashable, ...]:
        return (self.node,)

    def as_node(self) -> dm.NodeApply:
        if self.usage == "type":
            return dm.NodeApply(space=self.node.space, external_id=self.node.external_id)
        elif self.usage == "collection":
            raise NotImplementedError("Collection nodes are not supported yet")
        else:
            raise ValueError(f"Unknown usage {self.usage}")

    @field_serializer("node", when_used="unless-none")
    def remove_default_space(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, DMSEntity) and (metadata := _metadata(info.context)):
            return value.dump(space=metadata.space, version=metadata.version)
        return str(value)


class DMSEnum(SheetRow):
    collection: ClassEntityType = Field(alias="Collection")
    value: str = Field(alias="Value")
    name: str | None = Field(alias="Name", default=None)
    description: str | None = Field(alias="Description", default=None)

    def _identifier(self) -> tuple[Hashable, ...]:
        return self.collection, self.value

    @field_serializer("collection", when_used="unless-none")
    def remove_default_space(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, DMSEntity) and (metadata := _metadata(info.context)):
            return value.dump(space=metadata.space, version=metadata.version)
        return str(value)


class DMSRules(BaseRules):
    metadata: DMSMetadata = Field(alias="Metadata")
    properties: SheetList[DMSProperty] = Field(alias="Properties")
    views: SheetList[DMSView] = Field(alias="Views")
    containers: SheetList[DMSContainer] | None = Field(None, alias="Containers")
    enum: SheetList[DMSEnum] | None = Field(None, alias="Enum")
    nodes: SheetList[DMSNode] | None = Field(None, alias="Nodes")
    last: "DMSRules | None" = Field(None, alias="Last", description="The previous version of the data model")
    reference: "DMSRules | None" = Field(None, alias="Reference")

    @field_validator("reference")
    def check_reference_of_reference(cls, value: "DMSRules | None", info: ValidationInfo) -> "DMSRules | None":
        if value is None:
            return None
        if value.reference is not None:
            raise ValueError("Reference rules cannot have a reference")
        if value.metadata.data_model_type == DataModelType.solution and (metadata := info.data.get("metadata")):
            warnings.warn(
                PrincipleSolutionBuildsOnEnterpriseWarning(
                    f"The solution model {metadata.as_data_model_id()} is referencing another "
                    f"solution model {value.metadata.as_data_model_id()}",
                ),
                stacklevel=2,
            )
        return value

    @field_validator("views")
    def matching_version_and_space(cls, value: SheetList[DMSView], info: ValidationInfo) -> SheetList[DMSView]:
        if not (metadata := info.data.get("metadata")):
            return value
        model_version = metadata.version
        if different_version := [view.view.as_id() for view in value if view.view.version != model_version]:
            for view_id in different_version:
                warnings.warn(
                    PrincipleMatchingSpaceAndVersionWarning(
                        f"The view {view_id!r} has a different version than the data model version, {model_version}",
                    ),
                    stacklevel=2,
                )
        if different_space := [view.view.as_id() for view in value if view.view.space != metadata.space]:
            for view_id in different_space:
                warnings.warn(
                    PrincipleMatchingSpaceAndVersionWarning(
                        f"The view {view_id!r} is in a different space than the data model space, {metadata.space}",
                    ),
                    stacklevel=2,
                )
        return value

    @model_validator(mode="after")
    def post_validation(self) -> "DMSRules":
        from ._validation import DMSPostValidation

        issue_list = DMSPostValidation(self).validate()
        if issue_list.warnings:
            issue_list.trigger_warnings()
        if issue_list.has_errors:
            raise MultiValueError(issue_list.errors)
        return self

    def as_schema(self, include_pipeline: bool = False, instance_space: str | None = None) -> DMSSchema:
        from ._exporter import _DMSExporter

        return _DMSExporter(self, include_pipeline, instance_space).to_schema()

    def _repr_html_(self) -> str:
        summary = {
            "type": "Physical Data Model",
            "intended for": "DMS Architect",
            "name": self.metadata.name,
            "space": self.metadata.space,
            "external_id": self.metadata.external_id,
            "version": self.metadata.version,
            "views": len(self.views),
            "containers": len(self.containers) if self.containers else 0,
            "properties": len(self.properties),
        }

        return pd.DataFrame([summary]).T.rename(columns={0: ""})._repr_html_()  # type: ignore



import typing
from abc import ABC, abstractmethod
from typing import ClassVar, TypeVar

from pydantic import BaseModel, ConfigDict

from cognite.neat._app.monitoring.metrics import NeatMetricsCollector
from cognite.neat._config import Config
from cognite.neat._workflows.model import FlowMessage, WorkflowConfigs


class Configurable(BaseModel):
    name: str
    value: str | None = None
    label: str | None = None
    type: str | None = None  # string , secret , number , boolean , json , multi_select , single_select
    required: bool = False
    options: list[str] | None = None


class DataContract(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)


T_Input = TypeVar("T_Input", bound=DataContract)
T_Input1 = TypeVar("T_Input1", bound=DataContract)
T_Input2 = TypeVar("T_Input2", bound=DataContract)
T_Output = TypeVar("T_Output", bound=DataContract)


class Step(ABC):
    description: str = ""
    category: str = "default"
    configurables: ClassVar[list[Configurable]] = []
    scope: str = "core_global"
    metrics: NeatMetricsCollector | None = None
    workflow_configs: WorkflowConfigs | None = None
    version: str = "1.0.0"  # version of the step. All alpha versions considered as experimental
    source: str = (
        "cognite"  # source of the step , can be source identifier or url , for instance github url for instance.
    )
    docs_url: str = "https://cognite-neat.readthedocs-hosted.com/en/latest/"  # url to the documentation of the step

    def __init__(self, config: Config):
        self.log: bool = False
        self.configs: dict[str, str] = {}
        self.complex_configs: dict[
            str, typing.Any
        ] = {}  # complex configs are meant for more complex configurations. Value can be any type.
        self.workflow_id: str = ""
        self.workflow_run_id: str = ""
        self.config = config
        self.data_store_path = config.data_store_path

    @property
    def _not_configured_message(self) -> str:
        return f"Step {type(self).__name__} has not been configured."

    def set_metrics(self, metrics: NeatMetricsCollector | None):
        self.metrics = metrics

    def set_workflow_configs(self, configs: WorkflowConfigs | None):
        self.workflow_configs = configs

    def set_workflow_metadata(self, workflow_id: str, workflow_run_id: str):
        self.workflow_id = workflow_id
        self.workflow_run_id = workflow_run_id

    def configure(self, configs: dict[str, str], complex_configs: dict[str, typing.Any] | None = None):
        if complex_configs is None:
            complex_configs = {}
        self.configs = configs
        self.complex_configs = complex_configs

    def set_flow_context(self, context: dict[str, DataContract]):
        self.flow_context = context

    @abstractmethod
    def run(self, *input_data: DataContract) -> DataContract | tuple[FlowMessage, DataContract] | FlowMessage: ...



from pathlib import Path
from typing import Any

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import DataModelIdentifier

from cognite.neat._graph import examples as instances_examples
from cognite.neat._graph import extractors
from cognite.neat._issues import IssueList
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._rules import importers
from cognite.neat._rules._shared import ReadRules

from ._state import SessionState
from ._wizard import NeatObjectType, RDFFileType, object_wizard, rdf_dm_wizard
from .exceptions import intercept_session_exceptions


@intercept_session_exceptions
class ReadAPI:
    def __init__(self, state: SessionState, client: CogniteClient | None, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose
        self.cdf = CDFReadAPI(state, client, verbose)
        self.rdf = RDFReadAPI(state, client, verbose)
        self.excel = ExcelReadAPI(state, client, verbose)


@intercept_session_exceptions
class BaseReadAPI:
    def __init__(self, state: SessionState, client: CogniteClient | None, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose
        self._client = client

    def _store_rules(self, io: Any, input_rules: ReadRules, source: str) -> None:
        if input_rules.rules:
            self._state.input_rules.append(input_rules)
            if self._verbose:
                if input_rules.issues.has_errors:
                    print(f"{source} {type(io)} {io} read failed")
                else:
                    print(f"{source} {type(io)} {io} read successfully")

    def _return_filepath(self, io: Any) -> Path:
        if isinstance(io, str):
            return Path(io)
        elif isinstance(io, Path):
            return io
        else:
            raise NeatValueError(f"Expected str or Path, got {type(io)}")


@intercept_session_exceptions
class CDFReadAPI(BaseReadAPI):
    def __init__(self, state: SessionState, client: CogniteClient | None, verbose: bool) -> None:
        super().__init__(state, client, verbose)
        self.classic = CDFClassicAPI(state, client, verbose)

    @property
    def _get_client(self) -> CogniteClient:
        if self._client is None:
            raise NeatValueError("No client provided. Please provide a client to read a data model.")
        return self._client

    def data_model(self, data_model_id: DataModelIdentifier) -> IssueList:
        importer = importers.DMSImporter.from_data_model_id(self._get_client, data_model_id)
        input_rules = importer.to_rules()
        self._store_rules(data_model_id, input_rules, "CDF")
        return input_rules.issues


@intercept_session_exceptions
class CDFClassicAPI(BaseReadAPI):
    @property
    def _get_client(self) -> CogniteClient:
        if self._client is None:
            raise ValueError("No client provided. Please provide a client to read a data model.")
        return self._client

    def assets(self, root_asset_external_id: str) -> None:
        extractor = extractors.AssetsExtractor.from_hierarchy(self._get_client, root_asset_external_id)
        self._state.store.write(extractor)
        if self._verbose:
            print(f"Asset hierarchy {root_asset_external_id} read successfully")


@intercept_session_exceptions
class ExcelReadAPI(BaseReadAPI):
    def __call__(self, io: Any) -> IssueList:
        filepath = self._return_filepath(io)
        input_rules: ReadRules = importers.ExcelImporter(filepath).to_rules()
        self._store_rules(io, input_rules, "Excel")
        return input_rules.issues


@intercept_session_exceptions
class RDFReadAPI(BaseReadAPI):
    def __init__(self, state: SessionState, client: CogniteClient | None, verbose: bool) -> None:
        super().__init__(state, client, verbose)
        self.examples = RDFExamples(state)

    def _ontology(self, io: Any) -> IssueList:
        filepath = self._return_filepath(io)
        input_rules: ReadRules = importers.OWLImporter.from_file(filepath).to_rules()
        self._store_rules(io, input_rules, "Ontology")
        return input_rules.issues

    def _imf(self, io: Any) -> IssueList:
        filepath = self._return_filepath(io)
        input_rules: ReadRules = importers.IMFImporter.from_file(filepath).to_rules()
        self._store_rules(io, input_rules, "IMF Types")
        return input_rules.issues

    def __call__(
        self,
        io: Any,
        type: NeatObjectType | None = None,
        source: RDFFileType | None = None,
    ) -> IssueList:
        if type is None:
            type = object_wizard()

        if type.lower() == "Data Model".lower():
            source = source or rdf_dm_wizard("What type of data model is the RDF?")
            if source == "Ontology":
                return self._ontology(io)
            elif source == "IMF":
                return self._imf(io)
            else:
                raise ValueError(f"Expected ontology, imf or instances, got {source}")

        elif type.lower() == "Instances".lower():
            self._state.store.write(extractors.RdfFileExtractor(self._return_filepath(io)))
            return IssueList()
        else:
            raise ValueError(f"Expected data model or instances, got {type}")


class RDFExamples:
    def __init__(self, state: SessionState) -> None:
        self._state = state

    @property
    def nordic44(self) -> IssueList:
        self._state.store.write(extractors.RdfFileExtractor(instances_examples.nordic44_knowledge_graph))
        return IssueList()



from cognite.neat._workflows.base import BaseWorkflow
from cognite.neat._workflows.manager import WorkflowManager
from cognite.neat._workflows.model import (
    FlowMessage,
    WorkflowFullStateReport,
    WorkflowStepDefinition,
    WorkflowStepEvent,
)

__all__ = [
    "BaseWorkflow",
    "WorkflowStepDefinition",
    "WorkflowManager",
    "WorkflowStepEvent",
    "WorkflowFullStateReport",
    "FlowMessage",
]



from abc import ABC
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any, Literal

from pydantic import ValidationError

from cognite.neat._issues import IssueList, MultiValueError, NeatError, NeatWarning, catch_warnings
from cognite.neat._issues.errors import NeatTypeError
from cognite.neat._rules._shared import (
    InputRules,
    MaybeRules,
    OutRules,
    ReadRules,
    T_InputRules,
    T_VerifiedRules,
    VerifiedRules,
)
from cognite.neat._rules.models import (
    AssetInputRules,
    AssetRules,
    DMSInputRules,
    DMSRules,
    DomainInputRules,
    DomainRules,
    InformationInputRules,
    InformationRules,
)

from ._base import RulesTransformer


class VerificationTransformer(RulesTransformer[T_InputRules, T_VerifiedRules], ABC):
    """Base class for all verification transformers."""

    _rules_cls: type[T_VerifiedRules]

    def __init__(self, errors: Literal["raise", "continue"]) -> None:
        self.errors = errors

    def transform(self, rules: T_InputRules | OutRules[T_InputRules]) -> MaybeRules[T_VerifiedRules]:
        issues = IssueList()
        in_: T_InputRules = self._to_rules(rules)
        error_args: dict[str, Any] = {}
        if isinstance(rules, ReadRules):
            error_args = rules.read_context
        verified_rules: T_VerifiedRules | None = None
        with _catch_issues(issues, NeatError, NeatWarning, error_args) as future:
            rules_cls = self._get_rules_cls(in_)
            verified_rules = rules_cls.model_validate(in_.dump())  # type: ignore[assignment]

        if (future.result == "failure" or issues.has_errors or verified_rules is None) and self.errors == "raise":
            raise issues.as_errors()
        return MaybeRules[T_VerifiedRules](
            rules=verified_rules,
            issues=issues,
        )

    def _get_rules_cls(self, in_: T_InputRules) -> type[T_VerifiedRules]:
        return self._rules_cls


class VerifyDMSRules(VerificationTransformer[DMSInputRules, DMSRules]):
    """Class to verify DMS rules."""

    _rules_cls = DMSRules


class VerifyInformationRules(VerificationTransformer[InformationInputRules, InformationRules]):
    """Class to verify Information rules."""

    _rules_cls = InformationRules


class VerifyAssetRules(VerificationTransformer[AssetInputRules, AssetRules]):
    """Class to verify Asset rules."""

    _rules_cls = AssetRules


class VerifyAnyRules(VerificationTransformer[InputRules, VerifiedRules]):
    """Class to verify arbitrary rules"""

    def _get_rules_cls(self, in_: InputRules) -> type[VerifiedRules]:
        if isinstance(in_, InformationInputRules):
            return InformationRules
        elif isinstance(in_, DMSInputRules):
            return DMSRules
        elif isinstance(in_, AssetInputRules):
            return AssetRules
        elif isinstance(in_, DomainInputRules):
            return DomainRules
        else:
            raise NeatTypeError(f"Unsupported rules type: {type(in_)}")


class _FutureResult:
    def __init__(self) -> None:
        self._result: Literal["success", "failure", "pending"] = "pending"

    @property
    def result(self) -> Literal["success", "failure", "pending"]:
        return self._result


@contextmanager
def _catch_issues(
    issues: IssueList,
    error_cls: type[NeatError] = NeatError,
    warning_cls: type[NeatWarning] = NeatWarning,
    error_args: dict[str, Any] | None = None,
) -> Iterator[_FutureResult]:
    """This is an internal help function to handle issues and warnings.

    Args:
        issues: The issues list to append to.
        error_cls: The class used to convert errors to issues.
        warning_cls:  The class used to convert warnings to issues.

    Returns:
        FutureResult: A future result object that can be used to check the result of the context manager.
    """
    with catch_warnings(issues, warning_cls):
        future_result = _FutureResult()
        try:
            yield future_result
        except ValidationError as e:
            issues.extend(error_cls.from_pydantic_errors(e.errors(), **(error_args or {})))
            future_result._result = "failure"
        except MultiValueError as e:
            issues.extend(e.errors)
            future_result._result = "failure"
        except NeatError as e:
            issues.append(e)
            future_result._result = "failure"
        else:
            future_result._result = "success"



from graphlib import CycleError
from typing import cast

from cognite.neat._issues import IssueList
from cognite.neat._issues.errors import NeatValueError, PropertyDefinitionError
from cognite.neat._rules.models._base_rules import SheetList
from cognite.neat._rules.models.asset._rules import AssetProperty, AssetRules
from cognite.neat._rules.models.entities import AssetEntity, AssetFields, ClassEntity
from cognite.neat._rules.models.information._validation import InformationPostValidation


class AssetPostValidation(InformationPostValidation):
    def validate(self) -> IssueList:
        self.issue_list = super().validate()
        self._parent_property_point_to_class()
        self._circular_dependency()
        return self.issue_list

    def _parent_property_point_to_class(self) -> None:
        for property_ in cast(SheetList[AssetProperty], self.properties):
            for implementation in property_.implementation:
                if (
                    isinstance(implementation, AssetEntity)
                    and implementation.property_ == AssetFields.parentExternalId
                    and not isinstance(property_.value_type, ClassEntity)
                ):
                    self.issue_list.append(
                        PropertyDefinitionError(
                            property_.class_,
                            "class",
                            property_.property_,
                            "parentExternalId is only allowed to "
                            f"point to a Class not {type(property_.value_type).__name__}",
                        )
                    )

    def _circular_dependency(self) -> None:
        from cognite.neat._rules.analysis import AssetAnalysis

        try:
            _ = AssetAnalysis(cast(AssetRules, self.rules)).class_topological_sort()
        except CycleError as error:
            self.issue_list.append(
                NeatValueError(f"Invalid Asset Hierarchy, circular dependency detected: {error.args[1]}")
            )



from dataclasses import dataclass
from typing import Generic

from cognite.neat._issues._base import ResourceType

from ._resources import ResourceNeatWarning, T_Identifier, T_ReferenceIdentifier


@dataclass(unsafe_hash=True)
class PropertyWarning(ResourceNeatWarning[T_Identifier]):
    """Base class for property warnings {resource_type} with identifier {identifier}.{property_name}"""

    property_name: str


@dataclass(unsafe_hash=True)
class PropertyTypeNotSupportedWarning(PropertyWarning[T_Identifier]):
    """The {resource_type} with identifier {identifier} has a property {property_name}
    of unsupported type {property_type}. This will be ignored."""

    property_type: str


@dataclass(unsafe_hash=True)
class PropertyNotFoundWarning(PropertyWarning, Generic[T_Identifier, T_ReferenceIdentifier]):
    """The {resource_type} with identifier {identifier} does not have a property {property_name} referred
    to by {referred_type} {referred_by} does not exist. This will be ignored.
    """

    fix = "Ensure the {resource_type} {identifier} has a {property_name} property"
    referred_by: T_ReferenceIdentifier
    referred_type: ResourceType


@dataclass(unsafe_hash=True)
class PropertyDefinitionDuplicatedWarning(PropertyWarning[T_Identifier]):
    """Got multiple values for the {resource_type} {identifier}.{property_name} {values}.
    {default_action}"""

    extra = "Recommended action: {recommended_action}"

    values: frozenset[str]
    default_action: str
    recommended_action: str | None = None


@dataclass(unsafe_hash=True)
class PropertyValueTypeUndefinedWarning(PropertyWarning[T_Identifier]):
    """The {resource_type} with identifier {identifier} has a property {property_name}
    which has undefined value type. This may result in unexpected behavior when exporting rules.
    {default_action}"""

    extra = "Recommended action: {recommended_action}"

    default_action: str
    recommended_action: str | None = None
